/*
 *  Copyright (C) 2007 Rob Cliff.  All Rights Reserved.
 *  Copyright (C) 2007 Savage Minds Ltd.  All Rights Reserved.
 *
 *  This is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 *  USA.
 */

#include <AknApp.h>
#include <eikenv.h>
#include <utf.h>
#include "VNCViewerDocument.h"
#include "VNCViewerAppUi.h"
#include "BTExtNotifiers.h"
#include "VNCViewer.hrh"

// PIPS includes
#include <sys\types.h>
#include <sys\unistd.h>
#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <sys\mman.h>
#include <sys\shm.h>
#include <sys\sem.h>

// Defaults
const TUint16 KDefaultPort = 5900;

/**
The connection settings defaults.
*/
TConnectionSettings::TConnectionSettings()
: iPort(KDefaultPort),
  iViewOnly(EFalse),
  iShared(ETrue),
  iCursor(ETrue),
  iX11Cursor(ETrue),
  iFullScreen(ETrue),
  iDepth(0),
  iEncodings(EVncSettingEncodingAny),
  iCompression(5),
  iQuality(5),
  iJpeg(ETrue),
  iPrivacyWarning(ETrue),
  iConnectionWarning(ETrue)	
	{
	}

/**
Creates and constructs the document. This is called by
CVNCViewerApplication::CreateDocumentL() which in turn is called by the 
UI framework.

@param aApp The application
@return CVNCViewDocument* The new document
*/
CVNCViewerDocument* CVNCViewerDocument::NewL(CAknApplication& aApp)
	{
	CVNCViewerDocument* self = new (ELeave) CVNCViewerDocument(aApp);
	CleanupStack::PushL(self);
	self->ConstructL();
	CleanupStack::Pop(self);
	return self;
	}

/**
The constructor of the document class just passes the
supplied reference to the constructor initialization list.

@param aApp The application
*/
CVNCViewerDocument::CVNCViewerDocument(CAknApplication& aApp)
	: CAknDocument(aApp)
	{
	}

/**
The destructor of the document class.
*/
CVNCViewerDocument::~CVNCViewerDocument()
	{
	}

/**
2nd stage construction of the model.
All code that shall be called in initializing phase and might leave shall be 
added here.
*/
void CVNCViewerDocument::ConstructL()
	{
	// Read the INI file
	TRAPD(err, ReadIniFileL());

	// If there has been a problem, re-initialize the settings
	if (KErrNone != err)
		{
		SaveIniFileL();
		}
	}

/**
This is called by the UI framework as soon as the document has been created. 
It creates an instance of the ApplicationUI. The Application UI class is an 
instance of a CEikAppUi derived class.

@return CEikAppUi* The new App Ui
*/
CEikAppUi* CVNCViewerDocument::CreateAppUiL()
	{
	return new (ELeave) CVNCViewerAppUi;
	}

/**
This is called at startup to read the previous settings.
*/
void CVNCViewerDocument::ReadIniFileL()
	{
	// Open the ini file
	CDictionaryStore* iniFile = Application()->OpenIniFileLC(CEikonEnv::Static()->FsSession());
	
	// Check whether the Ini has been initialized before
	if (iniFile->IsPresentL(Application()->AppDllUid()))
		{
		RDictionaryReadStream readStream;
		CleanupClosePushL(readStream);
		
		// Open the stream		
		readStream.OpenLC(*iniFile, Application()->AppDllUid());
		
		// Read the settings
		TPckg<TConnectionSettings> settings(iSettings);
		readStream.ReadL(settings);

		CleanupStack::PopAndDestroy(3); // readStream * 2, iniFile
		
		iSettings.iDepth = 0;
		}
	else
		{
		// The ini file has not been initialized
		CleanupStack::PopAndDestroy(iniFile);
		SaveIniFileL();
		}
	}

/**
This is called when a connection attempt is started to save the settings
*/
void CVNCViewerDocument::SaveIniFileL()
	{
	// Open the ini file
	CDictionaryStore* iniFile = Application()->OpenIniFileLC(CEikonEnv::Static()->FsSession());

	RDictionaryWriteStream writeStream;
	CleanupClosePushL(writeStream);

	// Open the stream	
	writeStream.AssignLC(*iniFile, Application()->AppDllUid());

	// Write the settings
	TPckg<TConnectionSettings> settings(iSettings);
	writeStream.WriteL(settings);

	// Commit the changes
	writeStream.CommitL();
	
	CleanupStack::PopAndDestroy(2); // writeStream * 2

	iniFile->Commit();
		
	CleanupStack::PopAndDestroy(iniFile);	
	}

/**
Retrieve the settings

@return TConnectionSettings& The settings
*/
TConnectionSettings& CVNCViewerDocument::Settings()
	{
	return iSettings;
	}
		
/**
Convert the settings structure to a command line
*/
void CVNCViewerDocument::SetupCmdParamsL()
	{
	iCmdParams.SetLength(0);

	if (iSettings.iViewOnly)
		iCmdParams.Append(_L("-viewonly "));
	
	if (iSettings.iShared)
		iCmdParams.Append(_L("-shared "));
	else	
		iCmdParams.Append(_L("-noshared "));
	
	if (!iSettings.iCursor)
		iCmdParams.Append(_L("-nocursorshape "));
	
	if (iSettings.iX11Cursor)
		iCmdParams.Append(_L("-x11cursor "));

	if (iSettings.iDepth != EVncSettingDepthAny)
		{
		switch(iSettings.iDepth)
			{
			case EVncSettingDepth32:
				iCmdParams.Append(_L(" -depth 32 -truecolour "));
				break;			
			case EVncSettingDepth16:
				iCmdParams.Append(_L(" -depth 16 -truecolour "));
				break;			
			case EVncSettingDepth8:
				iCmdParams.Append(_L(" -depth 16 -bgr233 "));		
				break;			
			}
		}

	switch(iSettings.iEncodings)
		{
		case EVncSettingEncodingTight:
			iCmdParams.Append(_L("-encodings tight "));
			break;
		case EVncSettingEncodingHextile:
			iCmdParams.Append(_L("-encodings hextile "));
			break;
		case EVncSettingEncodingZlib:
			iCmdParams.Append(_L("-encodings zlib "));
			break;
		case EVncSettingEncodingCoRRE:
			iCmdParams.Append(_L("-encodings corre "));
			break;
		case EVncSettingEncodingRRE:
			iCmdParams.Append(_L("-encodings rre "));
			break;
		case EVncSettingEncodingCopyRect:
			iCmdParams.Append(_L("-encodings copyrect "));
			break;
		case EVncSettingEncodingRaw:
			iCmdParams.Append(_L("-encodings raw "));
			break;
		default:
			break;			
		}

	if (iSettings.iCompression != 0)
		{
		iCmdParams.Append(_L("-compresslevel "));
		iCmdParams.AppendNum(iSettings.iCompression);
		iCmdParams.Append(' ');
		}

	if (iSettings.iQuality != 0)
		{
		iCmdParams.Append(_L("-quality "));
		iCmdParams.AppendNum(iSettings.iQuality);
		iCmdParams.Append(' ');
		}
	
	if (!iSettings.iJpeg)
		iCmdParams.Append(_L("-nojpeg "));
		
	if (iSettings.iVia.Length())
		{
		iCmdParams.Append(_L("-accesspoint \""));
		iCmdParams.Append(iSettings.iVia);
		iCmdParams.Append(_L("\" "));
		}
		
	iSettings.iAddress.TrimAll();
	iCmdParams.Append(iSettings.iAddress);
	iCmdParams.Append(_L("::"));
	iCmdParams.AppendNum(iSettings.iPort);	

	iCmdParams.ZeroTerminate();
	}
	
/**
Return the command line parameters

@return const TDesC& The command parameters
*/
const TDesC& CVNCViewerDocument::CmdParams()
	{
	return iCmdParams;
	}
	
