/*
 *  Copyright (C) 2007 Rob Cliff.  All Rights Reserved.
 *  Copyright (C) 2007 Savage Minds Ltd.  All Rights Reserved.
 *
 *  This is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 *  USA.
 */

#include "VNCViewerApplication.h"
#include "VNCViewerAppUi.h"
#include "VNCViewerVNCView.h"
#include "VNCViewerSettingsView.h"
#include "VNCViewerDocument.h"
#include "VNCViewer.hrh"
#include "VNCViewerExternalInterface.h"
#include <aknquerydialog.h>
#include <aknnotedialog.h>
#include <aknnotewrappers.h>
#include <utf.h>
#include <VNCViewer.rsg>

/**
Scroll defaults
*/
// The number of pixels at the edge of the view to trigger a bump
const TInt scrollBumpMargin = 10;

// The number of pixels to bump
const TInt scrollBump = 20;

// The number of pixels to move the cursor
const TInt moveCursor = 5;

// Scales
const TReal absoluteScale = 0.0;
const TReal fullscreenScale = 1.0;

/**
Creates and constructs the view.

@param aAppUi Reference to the AppUi
@return CVNCViewerVNCView* Pointer to a CVNCViewerVNCView object
*/
CVNCViewerVNCView* CVNCViewerVNCView::NewLC(CVNCViewerDocument& aDocument)
	{
	CVNCViewerVNCView* self = new (ELeave) CVNCViewerVNCView(aDocument);
	CleanupStack::PushL(self);
	self->ConstructL();
	return self;
	}

/**
Constructor for the view.
Passes the application UI reference to the construction of the super class.

KNullViewId should normally be passed as parent view for the applications
default view. The parent view is the logical view that is normally activated
when a go back command is issued. KNullViewId will activate the system
default view.

@param aDocument The document
*/
CVNCViewerVNCView::CVNCViewerVNCView(CVNCViewerDocument& aDocument)
	: CAknView(), iDocument(aDocument), iNeedToConnect(ETrue)
	{
	}

/**
Destructor for the view
*/
CVNCViewerVNCView::~CVNCViewerVNCView()
	{
	if (iVNCContainer)
		{
		AppUi()->RemoveFromViewStack(*this, iVNCContainer);
		delete iVNCContainer;
		iVNCContainer = NULL;
		}
	delete iAboutText;
	}

/**
2nd stage construction of the view.
*/
void CVNCViewerVNCView::ConstructL()
	{
	// Calls ConstructL that initialises the standard values.
	// This should always be called in the concrete view implementations.
	BaseConstructL(R_VNCVIEWER_VNCVIEW);
	iAboutText = CEikonEnv::Static()->AllocReadResourceAsDes16L(R_ABOUT_TEXT);
	}

/**
Returns the view Id

@return Returns the Uid of the view
*/
TUid CVNCViewerVNCView::Id() const
	{
	return KUidVNCViewerVNCView;
	}

/**
Change the appearance of a menu before it appears

@param aResourceId The menu pane ID
@param aMenuPane The menu pane class
*/
void CVNCViewerVNCView::DynInitMenuPaneL(TInt aResourceId, CEikMenuPane* aMenuPane)
	{
	switch(aResourceId)
		{
		case R_VNCVIEWER_VNCVIEW_MENU:
			{
			// Set the logged on state
			aMenuPane->SetItemDimmed(EVNCViewerSendText, !iVNCContainer->LoggedOn());
			aMenuPane->SetItemDimmed(EVNCViewerSendKey, !iVNCContainer->LoggedOn());
			aMenuPane->SetItemDimmed(EVNCViewerModifiers, !iVNCContainer->LoggedOn());
			aMenuPane->SetItemDimmed(EVNCViewerMouse, !iVNCContainer->LoggedOn());
			aMenuPane->SetItemDimmed(EVNCViewerScale, !iVNCContainer->LoggedOn());
			aMenuPane->SetItemDimmed(EVNCViewerCtrlAltDel, !iVNCContainer->LoggedOn());
			aMenuPane->SetItemDimmed(EVNCViewerRefresh, !iVNCContainer->LoggedOn());
			aMenuPane->SetItemDimmed(EVNCViewerFindCursor, !(iVNCContainer->LoggedOn() && iDocument.Settings().iCursor));
			aMenuPane->SetItemDimmed(EVNCViewerDirectionKeys, !iVNCContainer->LoggedOn());
			aMenuPane->SetItemDimmed(EVNCViewerBack, !iVNCContainer->LoggedOn());
			aMenuPane->SetItemDimmed(EAknSoftkeyExit, !iVNCContainer->LoggedOn());
			}
			break;
		case R_VNCVIEWER_VNCVIEW_MENU_MODIFIERS_CASCADE:
			{
			TInt modifiers = iVNCContainer->Modifiers();
			aMenuPane->SetItemButtonState(EVNCViewerModifierShift, modifiers & CVNCViewerContainer::EModShift ? EEikMenuItemSymbolOn : EEikMenuItemSymbolIndeterminate);
			aMenuPane->SetItemButtonState(EVNCViewerModifierCtrl, modifiers & CVNCViewerContainer::EModCtrl ? EEikMenuItemSymbolOn : EEikMenuItemSymbolIndeterminate);
			aMenuPane->SetItemButtonState(EVNCViewerModifierAlt, modifiers & CVNCViewerContainer::EModAlt ? EEikMenuItemSymbolOn : EEikMenuItemSymbolIndeterminate);
			aMenuPane->SetItemButtonState(EVNCViewerModifierFunc, modifiers & CVNCViewerContainer::EModFunc ? EEikMenuItemSymbolOn : EEikMenuItemSymbolIndeterminate);
			}
			break;
		case R_VNCVIEWER_VNCVIEW_MENU_MOUSE_CASCADE:
			{
			CVNCViewerContainer::TButtonMode buttonMode = iVNCContainer->ButtonMode();
			switch(buttonMode)
				{
				case CVNCViewerContainer::EButtonLeft:
					aMenuPane->SetItemButtonState(EVNCViewerMouseLeft, EEikMenuItemSymbolOn);
					break;
				case CVNCViewerContainer::EButtonMiddle:
					aMenuPane->SetItemButtonState(EVNCViewerMouseMiddle, EEikMenuItemSymbolOn);
					break;
				case CVNCViewerContainer::EButtonRight:
					aMenuPane->SetItemButtonState(EVNCViewerMouseRight, EEikMenuItemSymbolOn);
					break;
				}
			}
			break;
		case R_VNCVIEWER_VNCVIEW_MENU_SCALE_CASCADE:
			{
			TReal scale = iVNCContainer->Scale();
			if (scale == fullscreenScale)
				aMenuPane->SetItemButtonState(EVNCViewerScaleFullScreen, EEikMenuItemSymbolOn);
			else if (scale == 0.75)
				aMenuPane->SetItemButtonState(EVNCViewerScale3_4, EEikMenuItemSymbolOn);
			else if (scale == 0.5)
				aMenuPane->SetItemButtonState(EVNCViewerScale1_2, EEikMenuItemSymbolOn);
			else if (scale == 0.25)
				aMenuPane->SetItemButtonState(EVNCViewerScale1_4, EEikMenuItemSymbolOn);
			else if (scale == absoluteScale)
				aMenuPane->SetItemButtonState(EVNCViewerScaleAccurate, EEikMenuItemSymbolOn);
			else
				aMenuPane->SetItemButtonState(EVNCViewerScaleCustom, EEikMenuItemSymbolOn);
			}
			break;
		case R_VNCVIEWER_VNCVIEW_MENU_DIRECTIONKEYS_CASCADE:
			{
			if (iVNCContainer->ArrowKeysControlCursor())
				aMenuPane->SetItemButtonState(EVNCViewerDirectionKeysCursor, EEikMenuItemSymbolOn);
			else
				aMenuPane->SetItemButtonState(EVNCViewerDirectionKeysCaret, EEikMenuItemSymbolOn);
			}
		}
	}

/**
Handles all commands in the view.
Called by the UI framework when a command has been issued.
The command Ids are defined in the .hrh file.

@param aCommand The command to be executed
*/
void CVNCViewerVNCView::HandleCommandL(TInt aCommand)
	{
	if ((iExiting) || (!iVNCContainer))
		return;

	switch(aCommand)
		{
		case EVNCViewerScaleFullScreen:
			iVNCContainer->SetScaleL(fullscreenScale);
			break;
		case EVNCViewerScale3_4:
			iVNCContainer->SetScaleL(0.75);
			break;
		case EVNCViewerScale1_2:
			iVNCContainer->SetScaleL(0.5);
			break;
		case EVNCViewerScale1_4:
			iVNCContainer->SetScaleL(0.25);
			break;
		case EVNCViewerScaleAccurate:
			iVNCContainer->SetScaleL(absoluteScale);
			break;
		case EVNCViewerScaleCustom:
			iVNCContainer->SetCustomScaleL();
			break;
		case EVNCViewerSendText:
			iVNCContainer->SendTextDialogL();
			break;
		case EVNCViewerFindCursor:
			iVNCContainer->FindCursorL();
			break;
		case EVNCViewerModifierShift:
			{
			TInt modifiers = iVNCContainer->Modifiers();
			iVNCContainer->SetModifiers(CVNCViewerContainer::EModShift, (modifiers & CVNCViewerContainer::EModShift) > 0 ? (TBool)EFalse : (TBool)ETrue);
			}
			break;
		case EVNCViewerModifierCtrl:
			{
			TInt modifiers = iVNCContainer->Modifiers();
			iVNCContainer->SetModifiers(CVNCViewerContainer::EModCtrl, (modifiers & CVNCViewerContainer::EModCtrl) > 0 ? (TBool)EFalse : (TBool)ETrue);
			}
			break;
		case EVNCViewerModifierAlt:
			{
			TInt modifiers = iVNCContainer->Modifiers();
			iVNCContainer->SetModifiers(CVNCViewerContainer::EModAlt, (modifiers & CVNCViewerContainer::EModAlt) > 0 ? (TBool)EFalse : (TBool)ETrue);
			}
			break;
		case EVNCViewerModifierFunc:
			{
			TInt modifiers = iVNCContainer->Modifiers();
			iVNCContainer->SetModifiers(CVNCViewerContainer::EModFunc, (modifiers & CVNCViewerContainer::EModFunc) > 0 ? (TBool)EFalse : (TBool)ETrue);
			}
			break;
		case EVNCViewerMouseLeft:
			iVNCContainer->SetButtonMode(CVNCViewerContainer::EButtonLeft);
			break;
		case EVNCViewerMouseMiddle:
			iVNCContainer->SetButtonMode(CVNCViewerContainer::EButtonMiddle);
			break;
		case EVNCViewerMouseRight:
			iVNCContainer->SetButtonMode(CVNCViewerContainer::EButtonRight);
			break;
		case EVNCViewerRefresh:
			iVNCContainer->RefreshL();
			break;
		case EVNCViewerDirectionKeysCursor:
			iVNCContainer->SetArrowKeysControlCursor(ETrue);
			break;
		case EVNCViewerDirectionKeysCaret:
			iVNCContainer->SetArrowKeysControlCursor(EFalse);
			break;
		case EVNCViewerBack:
			iExiting = ETrue;
			iGoBack = ETrue;
			iVNCContainer->PostQuit();
			CActiveScheduler::Start();
			break;
		case EVNCViewerAbout:
			{
			CAknInformationNote* dialog = new(ELeave) CAknInformationNote();
			dialog->ExecuteLD(*iAboutText);
			}
			break;
		case EEikCmdExit:
		case EAknSoftkeyExit:
			iExiting = ETrue;
			iVNCContainer->PostQuit();
			CActiveScheduler::Start();
			AppUi()->Exit();
			break;
		// Go back and exit command will be passed to the CQikViewBase to handle.
		default:
			// Check if we are sending a key
			if (!iVNCContainer->SendKeyL(aCommand))
				AppUi()->HandleCommandL(aCommand);
			break;
		}
	}

/**
The view has been activated

@param aPrevViewId The previous view id
@param aCustomMessageId A message Uid
@param aCustomMessage A custom message
*/
void CVNCViewerVNCView::DoActivateL(const TVwsViewId& /* aPrevViewId */,
                         TUid /* aCustomMessageId */,
                         const TDesC8& /* aCustomMessage */)
	{
	// Connect for the first time only
	if (iNeedToConnect)
		{
		iGoBack = EFalse;
		iExiting = EFalse;

		iVNCContainer = new(ELeave) CVNCViewerContainer(*this, iDocument);

		// Create view container control that will then create
		// all child controls
		iVNCContainer->SetMopParent(this);

		// Add the container to the control stack
		// So that it and its children can receive key events
		AppUi()->AddToViewStackL(*this, iVNCContainer);

		// Hide the status and button bar
		StatusPane()->MakeVisible(EFalse);
		Cba()->MakeVisible(EFalse);

		// Complete the initialization
		iVNCContainer->ConstructL(ClientRect());

		iVNCContainer->ConnectL();
		iNeedToConnect = EFalse;
		}
	}

/**
 * View deactivation function intended for overriding by sub classes.
 * This function is called by @c AknViewDeactivated().
 */
void CVNCViewerVNCView::DoDeactivate()
	{
	}

/**
Forward on the foreground event

@param aForeground Foregrounnd or Background?
*/
void CVNCViewerVNCView::HandleForegroundEventL(TBool aForeground)
	{
	if ((iVNCContainer) && (iVNCContainer->LoggedOn()))
		iVNCContainer->HandleForegroundEventL(aForeground);
	}

/**
The UI configuration has changed
*/
void CVNCViewerVNCView::SizeChanged()
	{
	if (iVNCContainer)
		{
		iVNCContainer->SetRect(ClientRect());
		}
	}

/**
Force a quit
*/
void CVNCViewerVNCView::QuitNow()
	{
	CActiveScheduler::Stop();
	DestroyEngine(iGoBack);
	}

/**
The engine should be destroyed

@param aSwitchView Whether to switch to the settings view
*/
void CVNCViewerVNCView::DestroyEngine(TBool aSwitchView)
	{
	if (aSwitchView)
		{
		dynamic_cast<CVNCViewerAppUi*>(AppUi())->SetActiveView(KUidVNCViewerSettingsView);
		AppUi()->ActivateLocalViewL(KUidVNCViewerSettingsView);
		}

	AppUi()->RemoveFromViewStack(*this, iVNCContainer);
	delete iVNCContainer;
	iVNCContainer = NULL;
	iNeedToConnect = ETrue;

	// Show the status and button bar
	StatusPane()->MakeVisible(ETrue);
	Cba()->MakeVisible(ETrue);
	}

//---------------------------------------------------------------
//---------------------------------------------------------------



/**
Creates and constructs the viewer container

@param aView The VNC view
@param aDocument The document
*/
CVNCViewerContainer::CVNCViewerContainer(CVNCViewerVNCView& aView, CVNCViewerDocument& aDocument)
	: iView(aView), iDocument(aDocument), iArrowKeypressMovesPointer(ETrue), iPointerAcceleration(1), iCursorLocated(EFalse)
	{
	}

/**
2nd stage construction of the container.

@param aRect The bounding rectangle
*/
void CVNCViewerContainer::ConstructL(const TRect& aRect)
	{
	CreateWindowL();

    // Set the windows size
    SetRect(aRect);

    // Activate the window, which makes it ready to be drawn
    ActivateL();

    // Check for extended keyboard
    TInt machineId;
    HAL::Get(HAL::EMachineUid, machineId);
    if (machineId == 0x20001858)
    	iExtendedKeyboard = ETrue;
   	}

/**
Destructor for the container
*/
CVNCViewerContainer::~CVNCViewerContainer()
	{
	iCursor.Close();
	delete iCursorBitmap;
	delete iCursorMask;

	delete iVNCViewMsgProc;
	iVNCViewMsgProc = NULL;

	delete iMainBitmap;
	delete iSendTextActive;
	delete iScaleActive;
	}

/**
The number of component controls

@return TInt The number of component controls
*/
TInt CVNCViewerContainer::CountComponentControls() const
	{
	return 0;
	}

/**
Retrieve the sub controls

@param aIndex The control index
@return CCoeControl* The control
*/
CCoeControl* CVNCViewerContainer::ComponentControl(TInt /* aIndex */) const
	{
	return NULL;
	}

/**
Is the engine logged on

@return TBool Logged on?
*/
TBool CVNCViewerContainer::LoggedOn()
	{
	return iLoggedOn;
	}

/**
The current state of the modifiers

@return TInt The modifier flags
*/
TInt CVNCViewerContainer::Modifiers()
	{
	return iModifiers;
	}

/**
The current state of mouse button mode

@return TButtonMode The button mode
*/
CVNCViewerContainer::TButtonMode CVNCViewerContainer::ButtonMode()
	{
	return iButtonMode;
	}

/**
From MVNCViewerEvent, Show a password dialog

@return HBufC8* The password
*/
HBufC8* CVNCViewerContainer::ShowPasswordL()
	{
	TBuf<16> password;

	CAknTextQueryDialog* passwordDlg = new (ELeave) CAknTextQueryDialog(password);

	// Launch the dialog
	passwordDlg->ExecuteLD(R_VNCVIEWER_PASSWORD_DIALOG);

	return CnvUtfConverter::ConvertFromUnicodeToUtf8L(password);
	}


/**
From MVNCViewerEvent, Create the bitmap

@param aHandle The bitmap handle
*/
void CVNCViewerContainer::CreateBitmapL(TInt aHandle)
	{
	// Create the bitmap and activate it
	iMainBitmap = new(ELeave) CFbsBitmap();
	User::LeaveIfError(iMainBitmap->Duplicate(aHandle));

	TSize bitmapSize = iMainBitmap->SizeInPixels();

	// Send the graphics expose event
	iVNCViewMsgProc->SendGraphicsExposeEventL(0, 0, bitmapSize.iWidth, bitmapSize.iHeight);

	// We're now logged on
	iLoggedOn = ETrue;
	}

/**
From MVNCViewerEvent, Draw the bitmap onto the screen

@param aRect The bounding rectangle
*/
void CVNCViewerContainer::PutImage(const TRect& /* aRect */)
	{
	DrawNow();
	}

/**
From MVNCViewerEvent, The engine has ended
*/
void CVNCViewerContainer::EngineQuit()
	{
	iView.QuitNow();
	}

/**
From MVNCViewerEvent, The engine is moving the soft cursor

@param aX The cursor X position
@param aY The cursor Y position
*/
void CVNCViewerContainer::MoveSoftCursorL(TInt aX, TInt aY)
	{
	iCursorPos.iX = aX;
	iCursorPos.iY = aY;
	if (!iCursorLocated)
		{
		FindCursorL();
		iCursorLocated = ETrue;
		}

	DrawNow();
	}

/**
From MVNCViewerEvent, The engine is asking the UI to create a cursor

@param aBitmapHandle The handle to the cursor bitmap
@param aMaskHandle The handle to the cursor mask
@param aHotpointX The hot point X position on the cursor
@param aHotpointY The hot point Y position on the cursor
*/
RWsSprite* CVNCViewerContainer::CreatePixmapCursorL(TInt aBitmapHandle, TInt aMaskHandle, TInt aHotpointX, TInt aHotpointY)
	{
	// Construct two new bitmaps
	iCursorBitmap = new(ELeave) CFbsBitmap();
	iCursorMask = new(ELeave) CFbsBitmap();

	// Duplicate the engine bitmaps
	User::LeaveIfError(iCursorBitmap->Duplicate(aBitmapHandle));
	User::LeaveIfError(iCursorMask->Duplicate(aMaskHandle));
	
	// Get the window server
	RWsSession& wsSession = CEikonEnv::Static()->WsSession();

 	// Adjust pointer to local positioning
 	TPoint cursorPoint(iCursorPos.iX, iCursorPos.iY);
 	TranslateRemoteToLocalPoint(cursorPoint, ETrue);

	// The construct the pointer
	iCursor = RWsSprite(wsSession);
	User::LeaveIfError(iCursor.Construct(*DrawableWindow(), cursorPoint, 0));

	// Append the bitmap
	TSpriteMember member;
	member.iBitmap = iCursorBitmap;
	member.iMaskBitmap = iCursorMask;
	member.iInvertMask = EFalse;
	member.iInterval = 0;
	member.iOffset =TPoint(-aHotpointX, -aHotpointY);
	User::LeaveIfError(iCursor.AppendMember(member));

	// Activate the pixmap
	User::LeaveIfError(iCursor.Activate());

	iCursorCreated = ETrue;

	return &iCursor;
	}


/**
From MVNCViewerEvent, The engine is asking the UI to move the cursor

@param aX The cursor X position
@param aY The cursor Y position
*/
void CVNCViewerContainer::WarpPointerL(TInt aX, TInt aY)
	{
	iCursorPos.iX = aX;
	iCursorPos.iY = aY;

	TPoint cursorPoint(aX, aY);
	TranslateRemoteToLocalPoint(cursorPoint, ETrue);

	// Get the window server
	if (iCursorCreated)
		iCursor.SetPosition(cursorPoint);

	if (!iCursorLocated)
		{
		iCursorLocated = ETrue;
		FindCursorL();
		}
	}

/**
From MVNCViewerEvent, Free the cursor
*/
void CVNCViewerContainer::FreeCursorL()
	{
	if (iCursorCreated)
		iCursor.Close();
	}

/**
From MVNCViewerEvent, The server is sending a message

@param The server message
*/
void CVNCViewerContainer::ServerMessageL(const TDesC& aMessage)
	{
	CAknNoteDialog* serverMessage = new(ELeave)CAknNoteDialog(CAknNoteDialog::EWarningTone, CAknNoteDialog::ENoTimeout);
	serverMessage->SetTextL(aMessage);
	serverMessage->SetTextWrapping(ETrue);
	serverMessage->ExecuteDlgLD(R_VNCVIEWER_SERVER_MESSAGE_DIALOG);

	// We've had an error
	iView.DestroyEngine(ETrue);
	}

/**
From MSendText, The send text dialog has closed, send the data on

@param aErr The error code
@param aText The text to send
*/
void CVNCViewerContainer::SendTextL(TInt aErr, const TDesC& aText)
	{
	if (KErrNone == aErr)
		{
		iVNCViewMsgProc->SendTextL(aText);
		}

	delete iSendTextActive;
	iSendTextActive = NULL;
	}

/**
From MScale, The scale has changed

@param aErr The error code
@param aText The new scale factor
*/
void CVNCViewerContainer::ScaleL(TInt aErr, const TReal aScale)
	{
	if (KErrNone == aErr)
		{
		SetScaleL(aScale);
		}

	delete iScaleActive;
	iScaleActive = NULL;
	}

/**
Set the arrow key mode

@param aOn Arrow controls cursor?
*/
void CVNCViewerContainer::SetArrowKeysControlCursor(TBool aOn)
	{
	iArrowKeypressMovesPointer = aOn;
	}

/**
Open the send text dialog
*/
void CVNCViewerContainer::SendTextDialogL()
	{
	if (!iSendTextActive)
		iSendTextActive = CSendTextActive::NewL(*this);
	}

/**
Open the scale dialog
*/
void CVNCViewerContainer::SetCustomScaleL()
	{
	if (!iScaleActive)
		iScaleActive = CScaleActive::NewL(*this, iScale);
	}

/**
Set the Mouse button mode (left, middle or right)

@param aButtonMode The mouse button mode
*/
void CVNCViewerContainer::SetButtonMode(TButtonMode aButtonMode)
	{
	iButtonMode = aButtonMode;
	}

/**
Set the modifiers

@param aModifier The modifiers
@param aOn Whether to set the modifiers on or off
*/
void CVNCViewerContainer::SetModifiers(TInt aModifier, TBool aOn)
	{
	if (aOn)
		iModifiers |= aModifier;
	else
		iModifiers &= EModAll - aModifier;
	}

/**
Refresh the screen
*/
void CVNCViewerContainer::RefreshL()
	{
	iVNCViewMsgProc->RefreshL();
	}

/**
Set the scale factor

@param aScale The new scale
*/
void CVNCViewerContainer::SetScaleL(TReal aScale)
	{
	if (!iMainBitmap)
		return;

	// Get the current cursor point
	iScale = aScale;

	TRect viewPortRect(TPoint(0,0), iMainBitmap->SizeInPixels());
	if (iScale == fullscreenScale)
		{
		iVNCViewMsgProc->SetFullscreen(ETrue);
		}
	else
		{
		iVNCViewMsgProc->SetFullscreen(EFalse);
		}

	FindCursorL();

	// Redraw
	DrawNow();
	}

/**
Start the process of connecting to the VNC server
*/
void CVNCViewerContainer::ConnectL()
	{
	// Create the engine
	TRAPD(err, iDocument.SetupCmdParamsL());
	if (KErrNone == err)
		{
		TRAPD(err, iVNCViewMsgProc = CVNCViewMsgProcessor::NewL(*this, iDocument.CmdParams()));
		if (KErrNone != err)
			{
			CEikonEnv::Static()->InfoWinL(R_VNCVIEWER_NOENGINE_TITLE, R_VNCVIEWER_NOENGINE);
			}

		// Store the fullscreen
		iScale = iDocument.Settings().iFullScreen ? fullscreenScale : absoluteScale;
		}
	else
		{
		// We've had an error
		iView.DestroyEngine(ETrue);
		}
	}

/**
Forward on any foreground events

@param aForeground Foreground or Background?
*/
void CVNCViewerContainer::HandleForegroundEventL(TBool aForeground)
	{
	if (iLoggedOn)
		iVNCViewMsgProc->HandleForegroundEventL(aForeground);
	}

/**
Locate the cursor
*/
void CVNCViewerContainer::FindCursorL()
	{
	if (iDocument.Settings().iCursor)
		{
		TPoint originPoint(iCursorPos);

		// Get the window size
		TSize bitmapSize = iMainBitmap->SizeInPixels();
		TSize windowSize = WindowSize();

		originPoint.iX -= windowSize.iWidth / 2;
		originPoint.iY -= windowSize.iHeight / 2;
		if (originPoint.iX < 0)
			originPoint.iX = 0;
		if (originPoint.iY < 0)
			originPoint.iY = 0;
		if (originPoint.iX > bitmapSize.iWidth - windowSize.iWidth)
			originPoint.iX = bitmapSize.iWidth - windowSize.iWidth;
		if (originPoint.iY > bitmapSize.iHeight - windowSize.iHeight)
			originPoint.iY = bitmapSize.iHeight - windowSize.iHeight;

		iOrigin = originPoint;

		// Move the pointer
		WarpPointerL(iCursorPos.iX, iCursorPos.iY);

		DrawNow();
		}
	}

/**
Handle any pointer events

@param aPointerEvent The pointer event
*/
void CVNCViewerContainer::HandlePointerEventL(const TPointerEvent& aPointerEvent)
	{
	if (!iLoggedOn)
		return;

	TPointerEvent event = aPointerEvent;

	// Alter the pointer event for middle and right hand buttons
	if (iButtonMode == EButtonMiddle)
		{
		if (event.iType == TPointerEvent::EButton1Down)
			event.iType = TPointerEvent::EButton2Down;
		else if (event.iType == TPointerEvent::EButton1Up)
			event.iType = TPointerEvent::EButton2Up;
		}
	if (iButtonMode == EButtonRight)
		{
		if (event.iType == TPointerEvent::EButton1Down)
			event.iType = TPointerEvent::EButton3Down;
		else if (event.iType == TPointerEvent::EButton1Up)
			event.iType = TPointerEvent::EButton3Up;
		}

	TPoint localCursorPos(event.iPosition);
	TranslateLocalToRemotePoint(event.iPosition, ETrue);
	iVNCViewMsgProc->SendMouseEventL(event);

	if (iScale != fullscreenScale)
		{
		TRect windowRect(Rect());

		if ((aPointerEvent.iType == TPointerEvent::EDrag) || (aPointerEvent.iType == TPointerEvent::EMove))
			{
			// Check if the pointer is at the left hand edge
			if (localCursorPos.iX < scrollBumpMargin)
				{
				iOrigin.iX -= scrollBump;
				DrawNow();
				}

			// Check if the pointer is at the top edge
			if (localCursorPos.iY < scrollBumpMargin)
				{
				iOrigin.iY -= scrollBump;
				DrawNow();
				}

			// Check if the pointer is at the right hand edge
			if (localCursorPos.iX > (windowRect.Size().iWidth - scrollBumpMargin))
				{
				iOrigin.iX += scrollBump;
				DrawNow();
				}

			// Check if the pointer is at the bottom edge
			if (localCursorPos.iY > (windowRect.Size().iHeight - scrollBumpMargin))
				{
				iOrigin.iY += scrollBump;
				DrawNow();
				}
			}
		}
	}

/**
Handle any keyboard events

@param aKeyEvent The key event
@param aType The key event type
@return TKeyResponse Whether the event was used
*/
TKeyResponse CVNCViewerContainer::OfferKeyEventL(const TKeyEvent &aKeyEvent, TEventCode aType)
	{
	if (!iLoggedOn)
		return EKeyWasNotConsumed;

	TConnectionSettings& settings = iDocument.Settings();

	TSize viewportSize = Rect().Size();
	TSize bitmapSize = iMainBitmap->SizeInPixels();

	TInt keytype = 0;

	// Set the message type
	switch(aType)
		{
		case EEventKey:
			keytype = EKey;
			break;
		case EEventKeyUp:
			keytype = EKeyUp;
			break;
		case EEventKeyDown:
			keytype = EKeyDown;
			break;
		default:
			return EKeyWasNotConsumed;
		}

	// Calculate the state of the modifiers
	TInt modifiers = 0;

	if ((iModifiers & EModShift) || aKeyEvent.iModifiers & EModifierShift)
		modifiers |= EKeyModShift;
	if ((iModifiers & EModCtrl) || aKeyEvent.iModifiers & EModifierCtrl)
		modifiers |= EKeyModCtrl;
	if ((iModifiers & EModAlt) || aKeyEvent.iModifiers & EModifierAlt)
		modifiers |= EKeyModAlt;
	if ((iModifiers & EModFunc) || aKeyEvent.iModifiers & EModifierFunc)
		modifiers |= EKeyModFunc;

	if (aKeyEvent.iRepeats > 0)
		++iPointerAcceleration;
	else
		iPointerAcceleration = 0;

	TInt keyCode = aKeyEvent.iCode;

	TUint16 keyToSend = 0;

	switch(keyCode)
		{
		case EKeyBackspace:
			keyToSend = Keysyms::KKeysymBackSpace;
			break;
		case EKeyTab:
			keyToSend = Keysyms::KKeysymTab;
			break;
		case EKeyEnter:
			keyToSend = Keysyms::KKeysymReturn;
			break;
		case EKeyEscape:
			keyToSend = Keysyms::KKeysymEscape;
			break;
		case EKeyDelete:
			keyToSend = Keysyms::KKeysymDelete;
			break;
		case EKeyPrintScreen:
			keyToSend = Keysyms::KKeysymPrint;
			break;
		case EKeyHome:
			keyToSend = Keysyms::KKeysymHome;
			break;
		case EKeyEnd:
			keyToSend = Keysyms::KKeysymEnd;
			break;
		case EKeyPageUp:
			keyToSend = Keysyms::KKeysymPage_Up;
			break;
		case EKeyPageDown:
			keyToSend = Keysyms::KKeysymPage_Down;
			break;
		case EKeyInsert:
			keyToSend = Keysyms::KKeysymInsert;
			break;
		case EKeyLeftArrow:
			if ((!settings.iCursor) && (iScale != fullscreenScale))
				{
				iOrigin.iX -= scrollBump;
				if (iOrigin.iX < 0)
					iOrigin.iX = 0;
				return EKeyWasConsumed;
				}

			if (iArrowKeypressMovesPointer)
				{
				TBool scroll = CursorInScreenArea();
				iCursorPos.iX -= (moveCursor + iPointerAcceleration);
				if (iCursorPos.iX < 0)
					iCursorPos.iX = 0;
				MimicPointerMoveL(scroll);
				return EKeyWasConsumed;
				}
			else
				keyToSend = Keysyms::KKeysymLeft;
			break;
		case EKeyRightArrow:
			if ((!settings.iCursor) && (iScale != fullscreenScale))
				{
				iOrigin.iX += scrollBump;
				if (iOrigin.iX > (bitmapSize.iWidth - viewportSize.iWidth))
					iOrigin.iX = bitmapSize.iWidth - viewportSize.iWidth;
				return EKeyWasConsumed;
				}

			if (iArrowKeypressMovesPointer)
				{
				TBool scroll = CursorInScreenArea();
				iCursorPos.iX += (moveCursor + iPointerAcceleration);
				if (iCursorPos.iX > bitmapSize.iWidth - 1)
					iCursorPos.iX = bitmapSize.iWidth - 1;
				MimicPointerMoveL(scroll);
				return EKeyWasConsumed;
				}
			else
				keyToSend = Keysyms::KKeysymRight;
			break;
		case EKeyUpArrow:
			if ((!settings.iCursor) && (iScale != fullscreenScale))
				{
				iOrigin.iY -= scrollBump;
				if (iOrigin.iY < 0)
					iOrigin.iY = 0;
				return EKeyWasConsumed;
				}

			if (iArrowKeypressMovesPointer)
				{
				TBool scroll = CursorInScreenArea();
				iCursorPos.iY -= (moveCursor + iPointerAcceleration);
				if (iCursorPos.iY < 0)
					iCursorPos.iY = 0;
				MimicPointerMoveL(scroll);
				return EKeyWasConsumed;
				}
			else
				keyToSend = Keysyms::KKeysymUp;
			break;
		case EKeyDownArrow:
			if ((!settings.iCursor) && (iScale != fullscreenScale))
				{
				iOrigin.iY += scrollBump;
				if (iOrigin.iY > (bitmapSize.iHeight - viewportSize.iHeight))
					iOrigin.iY = bitmapSize.iHeight - viewportSize.iHeight;
				return EKeyWasConsumed;
				}

			if (iArrowKeypressMovesPointer)
				{
				TBool scroll = CursorInScreenArea();
				iCursorPos.iY += (moveCursor + iPointerAcceleration);
				if (iCursorPos.iY > bitmapSize.iHeight - 1)
					iCursorPos.iY = bitmapSize.iHeight - 1;
				MimicPointerMoveL(scroll);
				return EKeyWasConsumed;
				}
			else
				keyToSend = Keysyms::KKeysymDown;
			break;
		case EKeyDevice3:
			if (iArrowKeypressMovesPointer)
				{
				iPointerAcceleration = 0;
				if (modifiers & EKeyModCtrl)
					{
					if (!iPointerDown)
						MimicPointerDownUpL(ETrue);
					else
						MimicPointerDownUpL(EFalse);
					}
				else
					{
					MimicPointerDownUpL(ETrue);
					MimicPointerDownUpL(EFalse);
					}
				return EKeyWasConsumed;
				}
			else
				keyToSend = keyCode;
			break;
		case EKeyLeftShift:
			keyToSend = Keysyms::KKeysymShift_L;
			modifiers = 0;
			break;
		case EKeyRightShift:
			keyToSend = Keysyms::KKeysymShift_R;
			modifiers = 0;
			break;
		case EKeyLeftAlt:
			keyToSend = Keysyms::KKeysymAlt_L;
			modifiers = 0;
			break;
		case EKeyRightAlt:
			keyToSend = Keysyms::KKeysymAlt_R;
			modifiers = 0;
			break;
		case EKeyLeftCtrl:
			keyToSend = Keysyms::KKeysymControl_L;
			modifiers = 0;
			break;
		case EKeyRightCtrl:
			keyToSend = Keysyms::KKeysymControl_R;
			modifiers = 0;
			break;
		case EKeyLeftFunc:
			keyToSend = Keysyms::KKeysymMeta_L;
			modifiers = 0;
			break;
		case EKeyRightFunc:
			keyToSend = Keysyms::KKeysymMeta_R;
			modifiers = 0;
			break;
		case EKeyF1:
			keyToSend = Keysyms::KKeysymF1;
			break;
		case EKeyF2:
			keyToSend = Keysyms::KKeysymF2;
			break;
		case EKeyF3:
			keyToSend = Keysyms::KKeysymF3;
			break;
		case EKeyF4:
			keyToSend = Keysyms::KKeysymF4;
			break;
		case EKeyF5:
			keyToSend = Keysyms::KKeysymF5;
			break;
		case EKeyF6:
			keyToSend = Keysyms::KKeysymF6;
			break;
		case EKeyF7:
			keyToSend = Keysyms::KKeysymF7;
			break;
		case EKeyF8:
			keyToSend = Keysyms::KKeysymF8;
			break;
		case EKeyF9:
			keyToSend = Keysyms::KKeysymF9;
			break;
		case EKeyF10:
			keyToSend = Keysyms::KKeysymF10;
			break;
		case EKeyF11:
			keyToSend = Keysyms::KKeysymF11;
			break;
		case EKeyF12:
			keyToSend = Keysyms::KKeysymF12;
			break;
		case '1':
			keyToSend = (!iExtendedKeyboard || (aKeyEvent.iModifiers & EModifierFunc)) ? '1' : 'r';
			break;
		case '2':
			keyToSend = (!iExtendedKeyboard || (aKeyEvent.iModifiers & EModifierFunc)) ? '2' : 't';
			break;
		case '3':
			keyToSend = (!iExtendedKeyboard || (aKeyEvent.iModifiers & EModifierFunc)) ? '3' : 'y';
			break;
		case '*':
			keyToSend = (!iExtendedKeyboard || (aKeyEvent.iModifiers & EModifierFunc)) ? '*' : 'u';
			break;
		case '4':
			keyToSend = (!iExtendedKeyboard || (aKeyEvent.iModifiers & EModifierFunc)) ? '4' : 'f';
			break;
		case '5':
			keyToSend = (!iExtendedKeyboard || (aKeyEvent.iModifiers & EModifierFunc)) ? '5' : 'g';
			break;
		case '6':
			keyToSend = (!iExtendedKeyboard || (aKeyEvent.iModifiers & EModifierFunc)) ? '6' : 'h';
			break;
		case '#':
			keyToSend = (!iExtendedKeyboard || (aKeyEvent.iModifiers & EModifierFunc)) ? '#' : 'j';
			break;
		case '7':
			keyToSend = (!iExtendedKeyboard || (aKeyEvent.iModifiers & EModifierFunc)) ? '7' : 'v';
			break;
		case '8':
			keyToSend = (!iExtendedKeyboard || (aKeyEvent.iModifiers & EModifierFunc)) ? '8' : 'b';
			break;
		case '9':
			keyToSend = (!iExtendedKeyboard || (aKeyEvent.iModifiers & EModifierFunc)) ? '9' : 'n';
			break;
		case '0':
			keyToSend = (!iExtendedKeyboard || (aKeyEvent.iModifiers & EModifierFunc)) ? '0' : 'm';
			break;
		default:
			if((aKeyEvent.iCode >= 0x20 && aKeyEvent.iCode <= 0x7e))
				keyToSend = aKeyEvent.iCode;
			else
				return EKeyWasNotConsumed;
		}

	// Send the keypress
	iVNCViewMsgProc->SendKeyEventL(keytype, keyToSend, modifiers);

	return EKeyWasConsumed;
	}

/**
Test whether the cursor is visible

@return TBool Whether the cursor is visible
*/
TBool CVNCViewerContainer::CursorInScreenArea()
	{
	if (iScale == fullscreenScale)
		return ETrue;
	else
		{
		TSize windowSize = WindowSize();

		if ((iCursorPos.iX > iOrigin.iX) &&
			(iCursorPos.iX < iOrigin.iX + windowSize.iWidth) &&
			(iCursorPos.iY > iOrigin.iY) &&
			(iCursorPos.iY < iOrigin.iY + windowSize.iHeight))
			{
			return ETrue;
			}
		else
			{
			return EFalse;
			}
		}
	}

/**
Return the size of the viewable window

@return TSize The size
*/
TSize CVNCViewerContainer::WindowSize() const
	{
	if (!iMainBitmap)
		return TSize(1, 1);
	TSize windowSize = Rect().Size();

	if (iScale != absoluteScale)
		{
		TSize bitmapSize = iMainBitmap->SizeInPixels();
		windowSize.iWidth = static_cast<TInt>(bitmapSize.iWidth * iScale);
		windowSize.iHeight = static_cast<TInt>(bitmapSize.iHeight * iScale);
		}
	return windowSize;
	}

/**
Mimic a pointer move

@param aScroll Whether to scroll the window
*/
void CVNCViewerContainer::MimicPointerMoveL(TBool aScroll)
	{
	TPointerEvent pointerEvent;
	pointerEvent.iType = iPointerDown ? TPointerEvent::EDrag : TPointerEvent::EMove;
	pointerEvent.iPosition = iCursorPos;
	iVNCViewMsgProc->SendMouseEventL(pointerEvent);

	if ((iScale != fullscreenScale) && (aScroll))
		{
		TSize bitmapSize = iMainBitmap->SizeInPixels();

		TPoint localCursorPos(pointerEvent.iPosition);
		TranslateRemoteToLocalPoint(localCursorPos, ETrue);

		TRect viewPortRect(Rect());

		TSize windowSize = WindowSize();

		TInt scrollAmount = moveCursor + iPointerAcceleration;

		// Check if the pointer is at the left hand edge
		if (localCursorPos.iX < scrollBumpMargin)
			{
			iOrigin.iX -= scrollAmount;
			if (iOrigin.iX < 0)
				iOrigin.iX = 0;
			DrawNow();
			}

		// Check if the pointer is at the top edge
		if (localCursorPos.iY < scrollBumpMargin)
			{
			iOrigin.iY -= scrollAmount;
			if (iOrigin.iY < 0)
				iOrigin.iY = 0;
			DrawNow();
			}

		// Check if the pointer is at the right hand edge
		if (localCursorPos.iX > (viewPortRect.Size().iWidth - scrollBumpMargin))
			{
			iOrigin.iX += scrollAmount;

			if (iOrigin.iX > (bitmapSize.iWidth - windowSize.iWidth))
				iOrigin.iX = bitmapSize.iWidth - windowSize.iWidth;

			DrawNow();
			}

		// Check if the pointer is at the bottom edge
		if (localCursorPos.iY > (viewPortRect.Size().iHeight - scrollBumpMargin))
			{
			iOrigin.iY += scrollAmount;

			if (iOrigin.iY > (bitmapSize.iHeight - windowSize.iHeight))
				iOrigin.iY = bitmapSize.iHeight - windowSize.iHeight;

			DrawNow();
			}
		}
	}

/**
Mimic a pointer up or down event

@param aDown Pointer Down or Up Event?
*/
void CVNCViewerContainer::MimicPointerDownUpL(TBool aDown)
	{
	iPointerDown = aDown;

	TPointerEvent pointerEvent;

	switch(iButtonMode)
		{
		case EButtonLeft:
			pointerEvent.iType = aDown ? TPointerEvent::EButton1Down : TPointerEvent::EButton1Up;
			break;
		case EButtonMiddle:
			pointerEvent.iType = aDown ? TPointerEvent::EButton2Down : TPointerEvent::EButton2Up;
			break;
		case EButtonRight:
			pointerEvent.iType = aDown ? TPointerEvent::EButton3Down : TPointerEvent::EButton3Up;
			break;
		}

	pointerEvent.iPosition.iX = iCursorPos.iX;
	pointerEvent.iPosition.iY = iCursorPos.iY;
	iVNCViewMsgProc->SendMouseEventL(pointerEvent);
	}

/**
Translate a remote (VNC server) coordinate to the local coordinate

@param aPoint The point to translate
@param aRelative Whether to take into account the relative coordinates of the top-left screen
*/
void CVNCViewerContainer::TranslateRemoteToLocalPoint(TPoint& aPoint, TBool aRelative) const
	{
	if (!iMainBitmap)
		return;
	TSize windowSize = Rect().Size();
	TSize bitmapSize = iMainBitmap->SizeInPixels();

	// aPoint contains the remote X & Y coords
	if (iScale == fullscreenScale)
		{
		// Fullscreen
		aPoint.iX = (aPoint.iX * windowSize.iWidth) / bitmapSize.iWidth;
		aPoint.iY = (aPoint.iY * windowSize.iHeight) / bitmapSize.iHeight;
		}
	else
		{
		if (aRelative)
			aPoint -= iOrigin;

		if (iScale != absoluteScale)
			{
			aPoint.iX = static_cast<TInt>(aPoint.iX * windowSize.iWidth / (iScale * bitmapSize.iWidth));
			aPoint.iY = static_cast<TInt>(aPoint.iY * windowSize.iHeight / (iScale * bitmapSize.iHeight));
			}
		}
	}

/**
Translate a local coordinate to the remote (VNC server) coordinate

@param aPoint The point to translate
@param aRelative Whether to take into account the relative coordinates of the top-left screen
*/
void CVNCViewerContainer::TranslateLocalToRemotePoint(TPoint& aPoint, TBool aRelative) const
	{
	if (!iMainBitmap)
		return;
	TSize bitmapSize = iMainBitmap->SizeInPixels();
	TSize windowSize = Rect().Size();

	// aPoint contains the remote X & Y coords
	if (iScale == fullscreenScale)
		{
		// Fullscreen
		aPoint.iX = (aPoint.iX * bitmapSize.iWidth) / windowSize.iWidth;
		aPoint.iY = (aPoint.iY * bitmapSize.iHeight) / windowSize.iHeight;
		}
	else
		{
		if (iScale != absoluteScale)
			{
			aPoint.iX = static_cast<TInt>(aPoint.iX * bitmapSize.iWidth * iScale / windowSize.iWidth);
			aPoint.iY = static_cast<TInt>(aPoint.iY * bitmapSize.iHeight * iScale / windowSize.iHeight);
			}

		if (aRelative)
			aPoint += iOrigin;
		}
	}

/**
Scale a rectangle using the current scale

@param aRect The unscaled rectangle
@return TRect The scaled rectangle
*/
TRect CVNCViewerContainer::ScaledRect(const TRect& aRect) const
	{
	if (iScale == absoluteScale)
		return aRect;

	TPoint rect_point(static_cast<TInt>(aRect.iTl.iX * iScale), static_cast<TInt>(aRect.iTl.iY * iScale));
	TSize rect_size(static_cast<TInt>(aRect.Width() * iScale), static_cast<TInt>(aRect.Height() * iScale));

	TRect scaledRect(rect_point, rect_size);
	return scaledRect;
	}

/**
Get the scaled viewable bitmap rectangle

@return TRect The scaled rectangle
*/
TRect CVNCViewerContainer::ViewRect() const
	{
	if (!iMainBitmap)
		return TRect(0, 0, 1, 1);
	TRect viewRect(TPoint(0, 0), iMainBitmap->SizeInPixels());
	return ScaledRect(viewRect);
	}

/**
Handle key events via a menu option

@param aKey The key to send
*/
TBool CVNCViewerContainer::SendKeyL(TInt aKey)
	{
	// Calculate the state of the modifiers
	TInt modifiers = 0;

	if (iModifiers & EModShift)
		modifiers |= EKeyModShift;
	if (iModifiers & EModCtrl)
		modifiers |= EKeyModCtrl;
	if (iModifiers & EModAlt)
		modifiers |= EKeyModAlt;
	if (iModifiers & EModFunc)
		modifiers |= EKeyModFunc;

	TInt keycode = 0;

	switch(aKey)
		{
		case EVNCViewerKeyEnter:
			keycode = Keysyms::KKeysymReturn;
			break;
		case EVNCViewerKeyBackspace:
			keycode = Keysyms::KKeysymBackSpace;
			break;
		case EVNCViewerKeyDelete:
			keycode = Keysyms::KKeysymDelete;
			break;
		case EVNCViewerKeyInsert:
			keycode = Keysyms::KKeysymInsert;
			break;
		case EVNCViewerKeyTab:
			keycode = Keysyms::KKeysymTab;
			break;
		case EVNCViewerKeyEscape:
			keycode = Keysyms::KKeysymEscape;
			break;
		case EVNCViewerKeyLeft:
			keycode = Keysyms::KKeysymLeft;
			break;
		case EVNCViewerKeyRight:
			keycode = Keysyms::KKeysymRight;
			break;
		case EVNCViewerKeyUp:
			keycode = Keysyms::KKeysymUp;
			break;
		case EVNCViewerKeyDown:
			keycode = Keysyms::KKeysymDown;
			break;
		case EVNCViewerKeyHome:
			keycode = Keysyms::KKeysymHome;
			break;
		case EVNCViewerKeyEnd:
			keycode = Keysyms::KKeysymEnd;
			break;
		case EVNCViewerKeyPgUp:
			keycode = Keysyms::KKeysymPage_Up;
			break;
		case EVNCViewerKeyPgDn:
			keycode = Keysyms::KKeysymPage_Down;
			break;
		case EVNCViewerKeyF1:
			keycode = Keysyms::KKeysymF1;
			break;
		case EVNCViewerKeyF2:
			keycode = Keysyms::KKeysymF2;
			break;
		case EVNCViewerKeyF3:
			keycode = Keysyms::KKeysymF3;
			break;
		case EVNCViewerKeyF4:
			keycode = Keysyms::KKeysymF4;
			break;
		case EVNCViewerKeyF5:
			keycode = Keysyms::KKeysymF5;
			break;
		case EVNCViewerKeyF6:
			keycode = Keysyms::KKeysymF6;
			break;
		case EVNCViewerKeyF7:
			keycode = Keysyms::KKeysymF7;
			break;
		case EVNCViewerKeyF8:
			keycode = Keysyms::KKeysymF8;
			break;
		case EVNCViewerKeyF9:
			keycode = Keysyms::KKeysymF9;
			break;
		case EVNCViewerKeyF10:
			keycode = Keysyms::KKeysymF10;
			break;
		case EVNCViewerKeyF11:
			keycode = Keysyms::KKeysymF11;
			break;
		case EVNCViewerKeyF12:
			keycode = Keysyms::KKeysymF12;
			break;
		case EVNCViewerKeyPrintScreen:
			keycode = Keysyms::KKeysymPrint;
			break;
		case EVNCViewerKeyPause:
			keycode = Keysyms::KKeysymPause;
			break;
		case EVNCViewerCtrlAltDel:
			keycode = Keysyms::KKeysymDelete;
			modifiers = EKeyModCtrl | EKeyModAlt;
			break;
		}

	// Send the keypress
	iVNCViewMsgProc->SendKeyEventL(EKey, keycode, modifiers);
	return (keycode != 0);
	}

/**
Post a quit message to the engine
*/
void CVNCViewerContainer::PostQuit()
	{
	iVNCViewMsgProc->PostQuit();
	}

/**
Draw the bitmap into the window

@param aRect The bounding rectangle
*/
void CVNCViewerContainer::Draw(const TRect& aRect) const
	{
	CWindowGc& gc = SystemGc();
  	gc.SetClippingRect(aRect);

	if (iLoggedOn)
		{
		if (iScale == fullscreenScale)
			{
			gc.DrawBitmap(Rect(), Bitmap());
			}
		else if (iScale == absoluteScale)
			{
			TRect sourceRect(iOrigin, aRect.Size());
			TPoint destPoint(0, 0);
			gc.BitBlt(destPoint, Bitmap(), sourceRect);
			}
		else
			{
			TRect viewRect(ViewRect());
			TPoint originPoint = iOrigin;

			// Work out the rect to show
			TRect rectToShow(originPoint, viewRect.Size());
			gc.DrawBitmap(Rect(), Bitmap(), rectToShow);
			}
		}
	else
		{
    	// Gets the control's extent
	    TRect drawRect(Rect());

    	// Clears the screen
		gc.Clear(drawRect);
		}

	gc.CancelClippingRect();
	}

