/*
 *  Copyright (C) 2007 Rob Cliff.  All Rights Reserved.
 *  Copyright (C) 2007 Savage Minds Ltd.  All Rights Reserved.
 *
 *  This is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 *  USA.
 */


#include <QikCommand.h>

#include "VNCViewerApplication.h"
#include "VNCViewerAppUi.h"
#include "VNCViewerSettingsView.h"
#include "VNCViewerDocument.h"
#include "VNCViewer.hrh"
#include "VNCViewerExternalInterface.h"
#include "VNCPasswordDialog.h"
#include "VNCViewAP.h"

#include <QikBuildingBlock.h>
#include <QikNumberEditor.h>
#include <QikVertOptionButtonList.h>
#include <EikEdwin.h>
#include <EIKCHLST.H>
#include <VNCViewer.rsg>

/**
Creates and constructs the settings view.

@param aAppUi Reference to the AppUi
@return CVNCViewerSettingsView* Pointer to a CVNCViewerSettingsView object
*/
CVNCViewerSettingsView* CVNCViewerSettingsView::NewLC(CQikAppUi& aAppUi)
	{
	CVNCViewerSettingsView* self = new (ELeave) CVNCViewerSettingsView(aAppUi);
	CleanupStack::PushL(self);
	self->ConstructL();
	return self;
	}

/**
Constructor for the view.
Passes the application UI reference to the construction of the super class.

KNullViewId should normally be passed as parent view for the applications 
default view. The parent view is the logical view that is normally activated 
when a go back command is issued. KNullViewId will activate the system 
default view. 

@param aAppUi Reference to the application UI
*/
CVNCViewerSettingsView::CVNCViewerSettingsView(CQikAppUi& aAppUi) 
	: CQikMultiPageViewBase(aAppUi, KNullViewId), iPrivacyDone(EFalse)
	{
	}

/**
Destructor for the view
*/
CVNCViewerSettingsView::~CVNCViewerSettingsView()
	{
	}

/**
2nd stage construction of the view.
*/
void CVNCViewerSettingsView::ConstructL()
	{
	// Calls ConstructL that initialises the standard values. 
	// This should always be called in the concrete view implementations.
	BaseConstructL();
	}
	
/**
Inherited from CQikViewBase and called upon by the UI Framework. 
It creates the view with commands from resource.
*/
void CVNCViewerSettingsView::ViewConstructL()
	{
	// Loads information about the UI configurations this view supports 
	// together with definition of each view, its layout and commands.
	ViewConstructFromResourceL(R_VNCVIEWER_UI_SETTINGS_CONFIGURATIONS);

	// Now set up all the defaults
	CVNCViewerDocument* document = static_cast<CVNCViewerDocument*>(iQikAppUi.Document());
	TConnectionSettings& settings = document->Settings();
	
	// Via
	CEikChoiceList* viaEditor = LocateControlByUniqueHandle<CEikChoiceList>(EConnectViaCtrl);

	// Create an array of access points
	CDesCArray* accessPoints = new(ELeave) CDesCArrayFlat(4);
	CleanupStack::PushL(accessPoints);
	CVNCViewAccessPoint::PopulateAccessPointsL(*accessPoints);
	
	TInt id = EVNCViewerPreferred;
	TInt selectedId = 0;
	CDesCArray* array = viaEditor->DesCArray();
	for (TInt count = 0; count < accessPoints->MdcaCount(); ++count)
		{	
		array->AppendL(accessPoints->MdcaPoint(count));
		if (settings.iVia.CompareF(accessPoints->MdcaPoint(count)) == 0)
			selectedId = count + 1;
		}
	CleanupStack::PopAndDestroy(accessPoints);
	viaEditor->SetCurrentItem(selectedId);
	
	// Address
	CEikEdwin* ipEditor = LocateControlByUniqueHandle<CEikEdwin>(EConnectAddressCtrl);
	ipEditor->SetTextL(&settings.iAddress);
	
	// Port
	CQikNumberEditor* portEditor = LocateControlByUniqueHandle<CQikNumberEditor>(EConnectPortCtrl);
	portEditor->SetValueL(settings.iPort);

	// Privacy
	CQikVertOptionButtonList* privacyEditor = LocateControlByUniqueHandle<CQikVertOptionButtonList>(EConnectPrivacyCtrl);
	privacyEditor->SetButtonById(settings.iPrivacyWarning ? EVNCYesOption : EVNCNoOption);

	// Billable
	CQikVertOptionButtonList* billableEditor = LocateControlByUniqueHandle<CQikVertOptionButtonList>(EConnectBillableCtrl);
	billableEditor->SetButtonById(settings.iConnectionWarning ? EVNCYesOption : EVNCNoOption);

	// ViewOnly
	CQikVertOptionButtonList* viewOnlyEditor = LocateControlByUniqueHandle<CQikVertOptionButtonList>(EControlViewOnlyCtrl);
	viewOnlyEditor->SetButtonById(settings.iViewOnly ? EVNCYesOption : EVNCNoOption);
	
	// Shared
	CQikVertOptionButtonList* sharedEditor = LocateControlByUniqueHandle<CQikVertOptionButtonList>(EControlSharedCtrl);
	sharedEditor->SetButtonById(settings.iShared ? EVNCYesOption : EVNCNoOption);

	// No cursor
	CQikVertOptionButtonList* cursorEditor = LocateControlByUniqueHandle<CQikVertOptionButtonList>(EControlCursorCtrl);
	cursorEditor->SetButtonById(settings.iCursor ? EVNCYesOption : EVNCNoOption);

	// X11 cursor
	CQikVertOptionButtonList* x11CursorEditor = LocateControlByUniqueHandle<CQikVertOptionButtonList>(EControlX11CursorCtrl);
	x11CursorEditor->SetButtonById(settings.iX11Cursor ? EVNCYesOption : EVNCNoOption);

	// Fullscreen
	CQikVertOptionButtonList* fullScreenEditor = LocateControlByUniqueHandle<CQikVertOptionButtonList>(EViewFullScreenCtrl);
	fullScreenEditor->SetButtonById(settings.iFullScreen ? EVNCYesOption : EVNCNoOption);

	// Depth	
	CQikVertOptionButtonList* depthEditor = LocateControlByUniqueHandle<CQikVertOptionButtonList>(EViewDepthCtrl);
	switch(settings.iDepth)
		{
		case 8:
			depthEditor->SetButtonById(EVNCDepth8);
			break;
		case 16:
			depthEditor->SetButtonById(EVNCDepth16);
			break;
		case 32:
			depthEditor->SetButtonById(EVNCDepth32);
			break;
		default:
			depthEditor->SetButtonById(EVNCDepthAny);
			break;
		}

	// Encoding
	CEikChoiceList* encodingsEditor = LocateControlByUniqueHandle<CEikChoiceList>(EEncodingEncodingsCtrl);
	encodingsEditor->SetCurrentItem(settings.iEncodings);

	// Jpeg
	CQikVertOptionButtonList* jpegEditor = LocateControlByUniqueHandle<CQikVertOptionButtonList>(EEncodingJpegCtrl);
	jpegEditor->SetButtonById(settings.iJpeg ? EVNCYesOption : EVNCNoOption);

	// Compression
	CQikNumberEditor* compressionEditor = LocateControlByUniqueHandle<CQikNumberEditor>(EEncodingCompressionCtrl);
	compressionEditor->SetValueL(settings.iCompression);

	// Quality
	CQikNumberEditor* qualityEditor = LocateControlByUniqueHandle<CQikNumberEditor>(EEncodingQualityCtrl);
	qualityEditor->SetValueL(settings.iQuality);
	
	// Now set up all the states
	TInt buttonId = cursorEditor->LabeledButtonId();
	LocateControlByUniqueHandle<CQikBuildingBlock>(EVNCViewerX11CursorBB)->SetDimmed(buttonId != EVNCYesOption);

	if ((settings.iEncodings == TConnectionSettings::EEncodingAny) || (settings.iEncodings == TConnectionSettings::EEncodingTight))
		{
		LocateControlByUniqueHandle<CQikBuildingBlock>(EVNCViewerJpegBB)->SetDimmed(EFalse);
		if (settings.iJpeg)
			{
			LocateControlByUniqueHandle<CQikBuildingBlock>(EVNCViewerCompressionBB)->SetDimmed(EFalse);
			LocateControlByUniqueHandle<CQikBuildingBlock>(EVNCViewerQualityBB)->SetDimmed(EFalse);
			}
		else
			{
			LocateControlByUniqueHandle<CQikBuildingBlock>(EVNCViewerCompressionBB)->SetDimmed(ETrue);
			LocateControlByUniqueHandle<CQikBuildingBlock>(EVNCViewerQualityBB)->SetDimmed(ETrue);
			}
		}
	else
		{
		LocateControlByUniqueHandle<CQikBuildingBlock>(EVNCViewerJpegBB)->SetDimmed(ETrue);
		LocateControlByUniqueHandle<CQikBuildingBlock>(EVNCViewerCompressionBB)->SetDimmed(ETrue);
		LocateControlByUniqueHandle<CQikBuildingBlock>(EVNCViewerQualityBB)->SetDimmed(ETrue);
		}
	}

/**
Returns the view Id

@return Returns the Uid of the view
*/
TVwsViewId CVNCViewerSettingsView::ViewId()const
	{
	return TVwsViewId(KUidVNCViewerApp, KUidVNCViewerSettingsView);
	}

/**
Handles all commands in the view.
Called by the UI framework when a command has been issued.
The command Ids are defined in the .hrh file.

@param aCommand The command to be executed
@see CQikViewBase::HandleCommandL
*/
void CVNCViewerSettingsView::HandleCommandL(CQikCommand& aCommand)
	{
	switch(aCommand.Id())
		{
		case EVNCViewerConnectCmd:
			{
			SaveSettingsL();
			
			CVNCViewerDocument* document = static_cast<CVNCViewerDocument*>(iQikAppUi.Document());
			TConnectionSettings& settings = document->Settings(); 
			if (settings.iConnectionWarning)
				{
				if (!CEikonEnv::Static()->QueryWinL(R_VNCVIEWER_BILLABLE_TITLE, R_VNCVIEWER_BILLABLE))
					{
					break;
					}
				}
			
			// Activates the vnc view
			dynamic_cast<CVNCViewerAppUi&>(iQikAppUi).SetActiveView(KUidVNCViewerVNCView);
			TVwsViewId viewId = TVwsViewId(KUidVNCViewerApp, KUidVNCViewerVNCView);
			iQikAppUi.ActivateViewL(viewId);
			break;
			}
		// Go back and exit command will be passed to the CQikViewBase to handle.
		default:
			CQikViewBase::HandleCommandL(aCommand);
			break;
		}
	}

/**
When the application needs to display a view, the view framework calls 
the view's ViewActivatedL() function. In some cases, the view is also passed 
a parameter that tells it what to display. In this view page1 is activated 
when the view gets visible.

@param aPrevViewId The composite view id of the previous view
@param aCustomMessageId The id of the message
@param aCustomMessage The message
*/
void CVNCViewerSettingsView::ViewActivatedL(const TVwsViewId& /*aPrevViewId*/, const TUid /*aCustomMessageId*/, const TDesC8& /*aCustomMessage*/)
	{
	if (!iPrivacyDone)
		{
		CVNCViewerDocument* document = static_cast<CVNCViewerDocument*>(iQikAppUi.Document());
		TConnectionSettings& settings = document->Settings();

		if (settings.iPrivacyWarning)
			{
			if (CEikonEnv::Static()->QueryWinL(R_VNCVIEWER_PRIVACY_TITLE, R_VNCVIEWER_PRIVACY))
				{
				CQikVertOptionButtonList* privacyEditor = LocateControlByUniqueHandle<CQikVertOptionButtonList>(EConnectPrivacyCtrl);
				privacyEditor->SetButtonById(EVNCNoOption);
				
				settings.iPrivacyWarning = EFalse;
				document->SaveIniFileL();
				}
			}
	
		iPrivacyDone = ETrue;
		}
		
	// View for page1 tab shall always be set as default when the view is 
	// activated.
	if (ActivePageId() != EVncViewerConnectPage)
		{
		ActivatePageL(EVncViewerConnectPage);
		}
	}

/**
Save the settings to the TConnectionSettings structure in the document
*/
void CVNCViewerSettingsView::SaveSettingsL()
	{
	// Now set up all the defaults
	CVNCViewerDocument* document = static_cast<CVNCViewerDocument*>(iQikAppUi.Document());
	TConnectionSettings& settings = document->Settings();

	// Via
	CEikChoiceList* viaEditor = LocateControlByUniqueHandle<CEikChoiceList>(EConnectViaCtrl);
	switch(viaEditor->CurrentItem())
		{
		case EVNCViewerPreferred:
			settings.iVia.SetLength(0);
			break;
		default:
			settings.iVia.Copy(viaEditor->Array()->MdcaPoint(viaEditor->CurrentItem()));
			break;
		}
	
	// Address
	CEikEdwin* ipEditor = LocateControlByUniqueHandle<CEikEdwin>(EConnectAddressCtrl);
	ipEditor->GetText(settings.iAddress);
	
	// Port
	CQikNumberEditor* portEditor = LocateControlByUniqueHandle<CQikNumberEditor>(EConnectPortCtrl);
	settings.iPort = portEditor->Value();

	// Privacy
	CQikVertOptionButtonList* privacyEditor = LocateControlByUniqueHandle<CQikVertOptionButtonList>(EConnectPrivacyCtrl);
	settings.iPrivacyWarning = (privacyEditor->LabeledButtonId() == EVNCYesOption);

	// Billable
	CQikVertOptionButtonList* billableEditor = LocateControlByUniqueHandle<CQikVertOptionButtonList>(EConnectBillableCtrl);
	settings.iConnectionWarning = (billableEditor->LabeledButtonId() == EVNCYesOption);

	// ViewOnly
	CQikVertOptionButtonList* viewOnlyEditor = LocateControlByUniqueHandle<CQikVertOptionButtonList>(EControlViewOnlyCtrl);
	settings.iViewOnly = (viewOnlyEditor->LabeledButtonId() == EVNCYesOption);
	
	// Shared
	CQikVertOptionButtonList* sharedEditor = LocateControlByUniqueHandle<CQikVertOptionButtonList>(EControlSharedCtrl);
	settings.iShared = (sharedEditor->LabeledButtonId() == EVNCYesOption);

	// No cursor
	CQikVertOptionButtonList* cursorEditor = LocateControlByUniqueHandle<CQikVertOptionButtonList>(EControlCursorCtrl);
	settings.iCursor = (cursorEditor->LabeledButtonId() == EVNCYesOption);

	// X11 cursor
	CQikVertOptionButtonList* x11CursorEditor = LocateControlByUniqueHandle<CQikVertOptionButtonList>(EControlX11CursorCtrl);
	settings.iX11Cursor = (x11CursorEditor->LabeledButtonId() == EVNCYesOption);

	// Fullscreen
	CQikVertOptionButtonList* fullScreenEditor = LocateControlByUniqueHandle<CQikVertOptionButtonList>(EViewFullScreenCtrl);
	settings.iFullScreen = (fullScreenEditor->LabeledButtonId() == EVNCYesOption);

	// Depth	
	CQikVertOptionButtonList* depthEditor = LocateControlByUniqueHandle<CQikVertOptionButtonList>(EViewDepthCtrl);
	TInt buttonId = depthEditor->LabeledButtonId();
	switch(buttonId)
		{
		case EVNCDepth8:
			settings.iDepth = 8;
			break;
		case EVNCDepth16:
			settings.iDepth = 16;
			break;
		case EVNCDepth32:
			settings.iDepth = 32;
			break;
		case EVNCDepthAny:
		default:
			settings.iDepth = 0;
			break;
		}

	// Encodings	
	CEikChoiceList* encodingsEditor = LocateControlByUniqueHandle<CEikChoiceList>(EEncodingEncodingsCtrl);
	settings.iEncodings = encodingsEditor->CurrentItem();

	// Compression
	CQikNumberEditor* compressionEditor = LocateControlByUniqueHandle<CQikNumberEditor>(EEncodingCompressionCtrl);
	settings.iCompression = compressionEditor->Value();

	CQikNumberEditor* qualityEditor = LocateControlByUniqueHandle<CQikNumberEditor>(EEncodingQualityCtrl);
	settings.iQuality = qualityEditor->Value();

	// Jpeg
	CQikVertOptionButtonList* jpegEditor = LocateControlByUniqueHandle<CQikVertOptionButtonList>(EEncodingJpegCtrl);
	settings.iJpeg = (jpegEditor->LabeledButtonId() == EVNCYesOption);
	
	// Save the settings for the document
	document->SaveIniFileL();
	}

/**
An event happened
@param aControl The control in question
@param aEventType The event type
*/
void CVNCViewerSettingsView::HandleControlEventL(CCoeControl *aControl, TCoeEvent aEventType)
	{
	if (EEventStateChanged == aEventType)
		{
		// Now set up all the defaults
		CVNCViewerDocument* document = static_cast<CVNCViewerDocument*>(iQikAppUi.Document());
		TConnectionSettings& settings = document->Settings();
		
		CQikVertOptionButtonList* cursorEditor = LocateControlByUniqueHandle<CQikVertOptionButtonList>(EControlCursorCtrl);
		CEikChoiceList* encodingsEditor = LocateControlByUniqueHandle<CEikChoiceList>(EEncodingEncodingsCtrl);
		CQikVertOptionButtonList* jpegEditor = LocateControlByUniqueHandle<CQikVertOptionButtonList>(EEncodingJpegCtrl);

		if (aControl == cursorEditor)
			{
			TInt buttonId = cursorEditor->LabeledButtonId();
			LocateControlByUniqueHandle<CQikBuildingBlock>(EVNCViewerX11CursorBB)->SetDimmed(buttonId != EVNCYesOption);
			}

		if (aControl == encodingsEditor)
			{
			if ((encodingsEditor->CurrentItem() == TConnectionSettings::EEncodingAny) || (encodingsEditor->CurrentItem() == TConnectionSettings::EEncodingTight))
				{
				LocateControlByUniqueHandle<CQikBuildingBlock>(EVNCViewerJpegBB)->SetDimmed(EFalse);
				if (settings.iJpeg)
					{
					LocateControlByUniqueHandle<CQikBuildingBlock>(EVNCViewerCompressionBB)->SetDimmed(EFalse);
					LocateControlByUniqueHandle<CQikBuildingBlock>(EVNCViewerQualityBB)->SetDimmed(EFalse);
					}
				else
					{
					LocateControlByUniqueHandle<CQikBuildingBlock>(EVNCViewerCompressionBB)->SetDimmed(ETrue);
					LocateControlByUniqueHandle<CQikBuildingBlock>(EVNCViewerQualityBB)->SetDimmed(ETrue);
					}
				}
			else
				{
				LocateControlByUniqueHandle<CQikBuildingBlock>(EVNCViewerJpegBB)->SetDimmed(ETrue);
				LocateControlByUniqueHandle<CQikBuildingBlock>(EVNCViewerCompressionBB)->SetDimmed(ETrue);
				LocateControlByUniqueHandle<CQikBuildingBlock>(EVNCViewerQualityBB)->SetDimmed(ETrue);
				}
			}

		if (aControl == jpegEditor)
			{
			TInt buttonId = jpegEditor->LabeledButtonId();
			LocateControlByUniqueHandle<CQikBuildingBlock>(EVNCViewerCompressionBB)->SetDimmed(buttonId != EVNCYesOption);
			LocateControlByUniqueHandle<CQikBuildingBlock>(EVNCViewerQualityBB)->SetDimmed(buttonId != EVNCYesOption);
			}

		}
	
	CQikViewBase::HandleControlEventL(aControl, aEventType);
	}

