/*
 *  Copyright (C) 2007 Rob Cliff.  All Rights Reserved.
 *  Copyright (C) 2007 Savage Minds Ltd.  All Rights Reserved.
 *
 *  This is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 *  USA.
 */

#include "VNCViewMsg.h"
#include <W32STD.H>

const TInt KMaxServerMessage = 500;

/**
Constructor for the engine
Passes the callback Mixin interface and the command parameters the construction.

@param aViewerEvent Reference to the callback API
@param aCmdParams The command parameters for the engine
@return Pointer to a CVNCViewMsgProcessor object
*/
CVNCViewMsgProcessor* CVNCViewMsgProcessor::NewL(MVNCViewerEvent& aViewerEvent, const TDesC& aCmdParams)
	{
	CVNCViewMsgProcessor* engine = new(ELeave) CVNCViewMsgProcessor(aViewerEvent);
	CleanupStack::PushL(engine);
	engine->ConstructL(aCmdParams);
	CleanupStack::Pop(engine); // engine
	return engine;
	}

/**
Destructor for the engine
*/
CVNCViewMsgProcessor::~CVNCViewMsgProcessor()
	{	
	// Close the engine process and its pipes
	for(TInt count = 0; count < 3; ++count)
		if (iStdFds[count] != -1)
			close(iStdFds[count]);

	if (iEngFd != -1)
		close(iEngFd);
	
	delete iProcessEventQueue;
	delete iMsgWatcher;
	delete iVncIpc;
	iEventQueue.Close();
	iFs.Close();
	}
			
/**
Constructor for the engine
Passes the callback Mixin interface.

@param aViewerEvent Reference to the callback API
*/
CVNCViewMsgProcessor::CVNCViewMsgProcessor(MVNCViewerEvent& aViewerEvent)
: iViewerEvent(aViewerEvent), iEngFd(-1)
	{
	for(TInt count = 0; count < 3; ++count)
		iStdFds[count] = -1;
	}
	
/**
2nd stage construction of the engine.

@param aCmdParams The command parameters for the engine
*/
void CVNCViewMsgProcessor::ConstructL(const TDesC& aCmdParams)
	{
	User::LeaveIfError(iFs.Connect());

	// Create the engine
	CreateEngineL(aCmdParams);

	// Create the ipc mechanism	
	iVncIpc = CVncIpc::NewL(ETrue);
					
	// Create the queue processor
	TCallBack callback(CVNCViewMsgProcessor::Event, this);
	iProcessEventQueue = new(ELeave) CAsyncCallBack(callback, CAsyncCallBack::EPriorityUserInput);

	// Start the message watcher
	iMsgWatcher = CVNCMessageWatcher::NewL(*iVncIpc, *this, iStdFds[2]);
	
	iMsgWatcher->Watch();
	}

/**
Called when an event is put into the event queue

@param aParam Callback param
@return Error code
*/
TInt CVNCViewMsgProcessor::Event(TAny* aParam)
	{
	CVNCViewMsgProcessor* engine = static_cast<CVNCViewMsgProcessor*>(aParam);
	TRAPD(err, engine->EventL());
	return KErrNone;
	}

/**
Protected call from the event processor.
*/
void CVNCViewMsgProcessor::EventL()
	{
	// Process any messages in the queue
	while (iEventQueue.Count())
		{
		struct vncmsg& msg = iEventQueue[0];
		TRAPD(err, iVncIpc->SendEventL(msg.iMsgType, msg));
		if (KErrNone != err)
			break;
		iEventQueue.Remove(0);
		}
	}
	
/**
A message is waiting
*/
void CVNCViewMsgProcessor::MsgWaitingL()
	{
	// Check if there are any messages in the event queue
	ProcessEventQueueReq();
	
	struct vncmsg msg;
	Mem::FillZ(&msg, sizeof(vncmsg));
	iVncIpc->RecvMessageL(msg);

	// Process the message
	switch(msg.iMsgType)
		{
		case EShowPasswordReq:
			ShowPasswordReqL();
			break;
		case ECreateBitmap:
			CreateBitmapL(msg);
			break;
		case EPutImage:
			PutImageL(msg);
			break;
		case ECopyClipboardToVNCAck:
			{
			TRAPD(err, CopyClipboardNowL());
			}
			break;
		case ECopyClipboardTransfer:
			{
			TRAPD(err, CopyIntoClipboardNowL(msg));
			}
			break;
		case EMoveSoftCursor:
			MoveSoftCursorL(msg);
			break; 		
		case ECreatePixmapCursorReq:
			CreatePixmapCursorL(msg);
			break;
		case EWarpPointer:
			WarpPointerL(msg);
			break;
		case EFreeCursor:
			FreeCursorL(msg);
			break;
		}

	// Watch for more messages
	iMsgWatcher->Watch();
	}

/**
An error message is waiting
*/
void CVNCViewMsgProcessor::StdErrWaitingL()
	{
	// Read the error
	TPtr8 error8 = HBufC8::NewLC(KMaxServerMessage)->Des();
	int length = read(iStdFds[2], (char*)error8.Ptr(), KMaxServerMessage);
	if (length > 0)
		{
		error8.SetLength(length);

		// Convert to UCS-2
		HBufC* bufError = CnvUtfConverter::ConvertToUnicodeFromUtf8L(error8); 
		CleanupStack::PushL(bufError);
		iViewerEvent.ServerMessageL(*bufError);
		CleanupStack::PopAndDestroy(bufError);
		}

	CleanupStack::PopAndDestroy(); // error8	
	}

/**
The engine has quit
*/
void CVNCViewMsgProcessor::QuitEngineL()
	{
	iViewerEvent.EngineQuit();	
	}

/**
Post a request to process the event queue if necessary
*/
void CVNCViewMsgProcessor::ProcessEventQueueReq()
	{
	if ((iEventQueue.Count() > 0) && (!iProcessEventQueue->IsActive()))
		iProcessEventQueue->Call();
	}

/**
Create the engine process

@param aCmdParams The command parameters
*/
void CVNCViewMsgProcessor::CreateEngineL(const TDesC& aCmdParams)
	{
	RProcess process;
	TFullName procName = process.FileName();
	process.Close();

	if (procName.Length() > 0)
		{
		procName.SetLength(1);
		_LIT(KExe, ":\\sys\\bin\\vncvieweng.exe\0");
		procName.Append(KExe);

		// Convert to UTF-8
		TPtr8 procName8(HBufC8::NewLC(procName.Length() * 3)->Des());
		CnvUtfConverter::ConvertFromUnicodeToUtf8(procName8, procName);
		procName8.ZeroTerminate();

		TPtr8 cmdParams8(HBufC8::NewLC(aCmdParams.Length() * 3)->Des());
		CnvUtfConverter::ConvertFromUnicodeToUtf8(cmdParams8, aCmdParams);
		cmdParams8.ZeroTerminate();

		iEngFd = popen3((char*)procName8.PtrZ(), (char*)cmdParams8.PtrZ(), NULL, iStdFds);
	
		if (iEngFd == -1)
			User::Leave(KErrNotFound);
		
		CleanupStack::PopAndDestroy(2); // procNameBuf8, cmdParamsBuf8
		}
	}

/**
Post a quit message to the engine
*/
void CVNCViewMsgProcessor::PostQuit()
	{
	struct vncmsg msg;
	QueueEvent(EQuitReq, msg);
	}

/**
Queue an event onto the message queue

@param aMsgType The message type of the message
@param aMsg The message structure
*/
void CVNCViewMsgProcessor::QueueEvent(TInt aMsgType, struct vncmsg& aMsg)
	{
	aMsg.iMsgType = aMsgType;
	iEventQueue.Append(aMsg);	
	ProcessEventQueueReq();
	}

/**
Received a password request from the engine
*/
void CVNCViewMsgProcessor::ShowPasswordReqL()
	{
	HBufC8* password = iViewerEvent.ShowPasswordL();
	TPtrC8 ptrPassword(*password);
	
	// Send the response
	struct vncmsg msg;
	Mem::FillZ(&msg, sizeof(vncmsg));
	strncpy(msg.iData.iPassword, (char*)ptrPassword.Ptr(), ptrPassword.Length());
	delete password;
	iVncIpc->SendMessageL(EShowPasswordResp, msg);
	}

/**
Received a request to create the bitmap

@param aMsg The message structure
*/
void CVNCViewMsgProcessor::CreateBitmapL(struct vncmsg& aMsg)
	{	
	// Create the bitmap
	TRAPD(err, iViewerEvent.CreateBitmapL(aMsg.iData.iCreateBitmap.iHandle));
	if (KErrNone != err)
		{
		User::Panic(_L("VNCViewer"), ECannotCreateBitmap);
		}
	}

/**
Received a request to copy the image to the screen

@param aMsg The message structure
*/
void CVNCViewMsgProcessor::PutImageL(struct vncmsg& aMsg)
	{
	TRect rect(aMsg.iData.iPutImage.iX, aMsg.iData.iPutImage.iX, 
		aMsg.iData.iPutImage.iWidth, aMsg.iData.iPutImage.iHeight);
	iViewerEvent.PutImage(rect);
	}
		
/**
Move a soft cursor

@param aMsg The message structure
*/
void CVNCViewMsgProcessor::MoveSoftCursorL(struct vncmsg& aMsg)
	{
	iViewerEvent.MoveSoftCursorL(aMsg.iData.iMoveSoftCursor.iX,
			aMsg.iData.iMoveSoftCursor.iY);
	}

/**
Create a cursor from two bitmaps

@param aMsg The message structure
*/
void CVNCViewMsgProcessor::CreatePixmapCursorL(struct vncmsg& aMsg)
	{
	RWsSprite* cursor = iViewerEvent.CreatePixmapCursorL(aMsg.iData.iCreatePixmapCursorReq.iPixmap,
			aMsg.iData.iCreatePixmapCursorReq.iMask,
			aMsg.iData.iCreatePixmapCursorReq.iHotPointX,
			aMsg.iData.iCreatePixmapCursorReq.iHotPointY);
			
	// Create the response
	struct vncmsg msg;
	msg.iData.iCreatePixmapCursorResp.iCursor = (TInt)cursor;
	iVncIpc->SendMessageL(ECreatePixmapCursorResp, msg);
	}

/**
Move the cursor

@param aMsg The message structure
*/
void CVNCViewMsgProcessor::WarpPointerL(struct vncmsg& aMsg)
	{
	iViewerEvent.WarpPointerL(aMsg.iData.iWarpPointer.iCursorX,
			aMsg.iData.iWarpPointer.iCursorY);
	}

/**
Free the cursor

@param aMsg The message structure
*/
void CVNCViewMsgProcessor::FreeCursorL(struct vncmsg& /* aMsg */)
	{
	iViewerEvent.FreeCursorL();
	}

/**
Send a graphics expose event to the VNC server

@param aX The x position of the expose
@param aY The y position of the expose
@param aWidth The width of the expose
@param aHeight The height of the expose
*/
void CVNCViewMsgProcessor::SendGraphicsExposeEventL(TInt aX, TInt aY, TInt aWidth, TInt aHeight)
	{
	struct vncmsg msg;
	msg.iData.iGraphicsExposeEvent.iX = aX;	
	msg.iData.iGraphicsExposeEvent.iY = aY;	
	msg.iData.iGraphicsExposeEvent.iWidth = aWidth;	
	msg.iData.iGraphicsExposeEvent.iHeight = aHeight;
	QueueEvent(EGraphicsExposeEvent, msg);
	}

/**
Send a mouse event to the VNC server

@param aEvent The pointer event
*/
void CVNCViewMsgProcessor::SendMouseEventL(const TPointerEvent& aEvent)
	{
	struct vncmsg msg;
	msg.iData.iMouseEvent.iX = aEvent.iPosition.iX;
	msg.iData.iMouseEvent.iY = aEvent.iPosition.iY;
	switch(aEvent.iType)
		{
		case TPointerEvent::EButton1Down:
			msg.iData.iMouseEvent.iType = EMouse1Down;
			break;
		case TPointerEvent::EButton1Up:
			msg.iData.iMouseEvent.iType = EMouse1Up;
			break;
		case TPointerEvent::EButton2Down:
			msg.iData.iMouseEvent.iType = EMouse2Down;
			break;
		case TPointerEvent::EButton2Up:
			msg.iData.iMouseEvent.iType = EMouse2Up;
			break;
		case TPointerEvent::EButton3Down:
			msg.iData.iMouseEvent.iType = EMouse3Down;
			break;
		case TPointerEvent::EButton3Up:
			msg.iData.iMouseEvent.iType = EMouse3Up;
			break;
		case TPointerEvent::EDrag:
			msg.iData.iMouseEvent.iType = EMouseDrag;
			break;
		case TPointerEvent::EMove:
			msg.iData.iMouseEvent.iType = EMouseMove;
			break;
		default:
			return;
		}
	
	QueueEvent(EMouseEvent, msg);
	}

/**
Send a selection of text to the VNC server

@param aText The text to send
*/
void CVNCViewMsgProcessor::SendTextL(const TDesC& aText)	
	{
	for(TInt count = 0; count < aText.Length(); ++count)
		{
		TUint16 key_char = aText[count];

		TBuf8<KMaxKeyLength> keychar;
		keychar.Num(key_char);
		keychar.ZeroTerminate();
		
		struct vncmsg msg;
		msg.iData.iKeyEvent.iType = EKey;
		strncpy(msg.iData.iKeyEvent.iKeyChar, (char*)keychar.Ptr(), KMaxKeyLength);
		QueueEvent(EKeyEvent, msg);		
		}
	}

/**
Send a key event to the VNC server, taking account of modifiers

@param aType The type of key event
@param aKeyChar The key code
@param aModifiers Any modifiers
*/
void CVNCViewMsgProcessor::SendKeyEventL(TInt aType, TUint16 aKeyChar, TInt aModifiers)
	{
	// Send any modifiers
	if (aModifiers & EKeyModShift)
		SendKeyEventL(EKeyDown, Keysyms::KKeysymShift_L);
	if (aModifiers & EKeyModCtrl)
		SendKeyEventL(EKeyDown, Keysyms::KKeysymControl_L);
	if (aModifiers & EKeyModAlt)
		SendKeyEventL(EKeyDown, Keysyms::KKeysymAlt_L);
	if (aModifiers & EKeyModFunc)
		SendKeyEventL(EKeyDown, Keysyms::KKeysymMeta_L);
	
	// Send the key
	SendKeyEventL(aType, aKeyChar);

	// Clear any modifiers
	if (aModifiers & EKeyModShift)
		SendKeyEventL(EKeyUp, Keysyms::KKeysymShift_L);
	if (aModifiers & EKeyModCtrl)
		SendKeyEventL(EKeyUp, Keysyms::KKeysymControl_L);
	if (aModifiers & EKeyModAlt)
		SendKeyEventL(EKeyUp, Keysyms::KKeysymAlt_L);
	if (aModifiers & EKeyModFunc)
		SendKeyEventL(EKeyUp, Keysyms::KKeysymMeta_L);
	}

/**
Send a key event to the VNC server

@params aType The type of key event
@params aKeyChar The key code
*/
void CVNCViewMsgProcessor::SendKeyEventL(TInt aType, TUint16 aKeyChar)
	{
	TBuf8<KMaxKeyLength> keychar;
	keychar.Num(aKeyChar);
	keychar.ZeroTerminate();

	struct vncmsg msg;
	msg.iData.iKeyEvent.iType = aType;
	strncpy(msg.iData.iKeyEvent.iKeyChar, (char*)keychar.Ptr(), KMaxKeyLength);
	QueueEvent(EKeyEvent, msg);		
	}
	
/**
Handle a foreground event, send the "Enter" or "Leave" event to the VNC server

@param aForeground The foreground or background
*/
void CVNCViewMsgProcessor::HandleForegroundEventL(TBool aForeground)
	{
	struct vncmsg msg;
	QueueEvent(aForeground ? ECopyClipboardToVNCReq : ECopyClipboardFromVNCReq, msg);		
	}

/**
Refresh the screen
*/
void CVNCViewMsgProcessor::RefreshL()
	{
	struct vncmsg msg;
	QueueEvent(ERefreshReq, msg);		
	}

/**
Copy the clipboard to the VNC server
*/
void CVNCViewMsgProcessor::CopyClipboardNowL()
	{
	HBufC8* buffer = 0;
	TRAPD(err, buffer = GetClipboardContentsL());	
	if (KErrNone == err)
		{
		CleanupStack::PushL(buffer);
		TPtrC8 ptrBuffer(*buffer);		

		// Send the request to the engine
		struct vncmsg msg;
		msg.iData.iCopyClipboard.iLength = ptrBuffer.Length();
		iVncIpc->SendMessageL(ECopyClipboardTransfer, msg, (char*)ptrBuffer.Ptr(), ptrBuffer.Length());
		
		CleanupStack::PopAndDestroy(buffer);
		}
	else
		{
		// Send the request to the engine
		struct vncmsg msg;
		msg.iData.iCopyClipboard.iLength = 0;
		iVncIpc->SendMessageL(ECopyClipboardTransfer, msg);
		}
	}

/**
Get the Symbian clipboard contents

@return HBufC8* The clipboard contents
*/
HBufC8* CVNCViewMsgProcessor::GetClipboardContentsL()
	{
	CClipboard* cb = CClipboard::NewForReadingLC(iFs);

	// Read into a plain text object
	CPlainText* plainText = CPlainText::NewL();
	CleanupStack::PushL(plainText);
	plainText->PasteFromStoreL(cb->Store(), cb->StreamDictionary(), 0);
	
	// Create a buffer to store the text
	TPtr8 ptrText = HBufC8::NewLC(plainText->DocumentLength() * 3)->Des();
	
	RDesWriteStream writeStream;
	CleanupClosePushL(writeStream);
	writeStream.Open(ptrText);
	
	plainText->ExportTextL(0, writeStream, CPlainText::EOrganiseByLine);

	TPtrC ptrText16((TUint16*)ptrText.Ptr(), ptrText.Size() / 2);
	
	// Convert to UTF8
	HBufC8* buf8 = CnvUtfConverter::ConvertFromUnicodeToUtf8L(ptrText16);	
		
	CleanupStack::PopAndDestroy(4); // buf8, writeStream, ptrText, plainText, cb

	return buf8;
	}

/**
Copy the clipboard from the VNC server

@param aMsg The message structure
*/
void CVNCViewMsgProcessor::CopyIntoClipboardNowL(struct vncmsg& aMsg)
	{
	TRAPD(err, CopyClipboardFromFifoL(aMsg.iData.iCopyClipboard.iLength));
	}
	
/**
Copy the clipboard from the VNC server 

@param aLength The length of data
*/
void CVNCViewMsgProcessor::CopyClipboardFromFifoL(TInt aLength)
	{
	TPtr8 ptr8(HBufC8::NewMaxLC(aLength)->Des());
	iVncIpc->RecvExtraData((char*)ptr8.Ptr(), aLength);

	// Convert from UTF8
	HBufC* buf = CnvUtfConverter::ConvertToUnicodeFromUtf8L(ptr8);
	TPtrC ptr(buf->Des());
	CleanupStack::PushL(buf);
	
	CClipboard* cb = CClipboard::NewForWritingLC(iFs);
	CPlainText* plainText = CPlainText::NewL();
	CleanupStack::PushL(plainText);
	
	plainText->InsertL(0, ptr);
	plainText->CopyToStoreL(cb->Store(), cb->StreamDictionary(), 0, ptr.Length());
	
	cb->CommitL();
	CleanupStack::PopAndDestroy(4); // plainText, cb, buf, ptr8
	}
	
/**
Set the full screen mode

@aOn Whether the fullscreen mode is on
*/
void CVNCViewMsgProcessor::SetFullscreen(TBool aOn)
	{
	struct vncmsg msg;
	msg.iData.iSetFullscreen.iFullscreen = aOn;
	iVncIpc->SendEventL(ESetFullscreen, msg);		
	}
	
//------------------------------------------------------------------
//------------------------------------------------------------------
		
/**
Constructor for the Message Watcher

@param aVncIpc The Ipc class
@param aMsgWaiting The message waiter callback
@param aStdErr The std err pipe
*/
CVNCMessageWatcher* CVNCMessageWatcher::NewL(CVncIpc& aVncIpc, MIpcMsgWaiting& aMsgWaiting, int aStdErr)
	{
	CVNCMessageWatcher* watcher = new(ELeave) CVNCMessageWatcher(aVncIpc, aMsgWaiting, aStdErr);
	CleanupStack::PushL(watcher);
	watcher->ConstructL();
	CleanupStack::Pop(watcher);
	return watcher;
	}
	
/**
Constructor for the Message Watcher

@param aVncIpc The Ipc class
@param aMsgWaiting The message waiter callback
@param aStdErr The std err pipe
*/
CVNCMessageWatcher::CVNCMessageWatcher(CVncIpc& aVncIpc, MIpcMsgWaiting& aMsgWaiting, int aStdErr)
: CActive(EPriorityStandard), iVncIpc(aVncIpc), iMsgWaiting(aMsgWaiting), iStdErr(aStdErr)
	{
	CActiveScheduler::Add(this);
	}

/**
2nd phase Constructor for the Message Watcher
*/
void CVNCMessageWatcher::ConstructL()
	{
	// Get the current thread id
	RThread thread;
	iThreadId = thread.Id();
	thread.Close();
	
	// Initialise the mutex and conditional variables
	pthread_mutex_init(&iMutex, NULL);
	pthread_cond_init(&iCond, NULL);
	pthread_cond_init(&iStartCond, NULL);
		
	// Start the thread
	pthread_create(&iWatchThread, NULL, CVNCMessageWatcher::WatchThread, this);

	// Wait until it is started
	pthread_mutex_lock(&iMutex);
	pthread_cond_wait(&iStartCond, &iMutex);
	pthread_mutex_unlock(&iMutex);
	}

/**
Destructor for the Message Watcher
*/
CVNCMessageWatcher::~CVNCMessageWatcher()
	{
	Cancel();
		
	pthread_mutex_destroy(&iMutex);
	pthread_cond_destroy(&iCond);	
	pthread_cond_destroy(&iStartCond);	
	}
	
/**
Watch the message queue
*/
void CVNCMessageWatcher::Watch()
	{
	pthread_mutex_lock(&iMutex);
	iStatus = KRequestPending;
	SetActive();
	pthread_cond_signal(&iCond);
	pthread_mutex_unlock(&iMutex);
	}

/**
The watcher thread

@param aArg The thread's arguments
*/
void* CVNCMessageWatcher::WatchThread(void* aArg)
	{
	CVNCMessageWatcher* watcher = (CVNCMessageWatcher*)aArg;
	watcher->WatchThread();
	return NULL;
	}
	
/**
The internal watcher thread
*/
void CVNCMessageWatcher::WatchThread()
	{
	// Open the UI thread, so we can signal it's iStatus
	RThread thread;
	thread.Open(iThreadId);
	
	// The thread has started
	pthread_cond_signal(&iStartCond);

	CVncIpc::TWaitPipe waitPipe;
	do
		{
		pthread_mutex_lock(&iMutex);

		// Wait until a watch command is issued
		pthread_cond_wait(&iCond, &iMutex);
		
		// Wait for data
		waitPipe = iVncIpc.WaitForData(iStdErr);
			
		// Signal that a pipe has data
		TRequestStatus* status = &iStatus;
		if (*status == KRequestPending)
			{
			if (waitPipe == CVncIpc::EWaitPipeNormal)
				thread.RequestComplete(status, KErrNone);
			if (waitPipe == CVncIpc::EWaitPipeStdErr)
				thread.RequestComplete(status, KErrArgument);
			if (waitPipe == CVncIpc::EWaitPipeQuit)
				thread.RequestComplete(status, KErrCompletion);
			}
		pthread_mutex_unlock(&iMutex);
		}
	while (waitPipe == CVncIpc::EWaitPipeNormal);
	
	// Close the thread resource
	thread.Close();
	}
	
/**
There is a message waiting for me, call back
*/
void CVNCMessageWatcher::RunL()
	{
	// Re-queue the watcher
	if (iStatus.Int() == KErrNone)
		{
		// Callback
		iMsgWaiting.MsgWaitingL();
		}
	else if (iStatus.Int() == KErrArgument)
		{
		// Callback
		iMsgWaiting.StdErrWaitingL();
		}
	else if (iStatus.Int() == KErrCompletion)
		{
		// Callback
		iMsgWaiting.QuitEngineL();
		}
	else
		{
		// Panic
		User::Panic(_L("VNCViewer_Watcher"), EUnexpected);		
		}
	}
	
/**
Cancel the watch
*/
void CVNCMessageWatcher::DoCancel()
	{
	TRequestStatus* status = &iStatus;
	User::RequestComplete(status, KErrCancel);
	}

	
	

	