/*
 *  Copyright (C) 2007 Rob Cliff.  All Rights Reserved.
 *  Copyright (C) 2007 Savage Minds Ltd.  All Rights Reserved.
 *
 *  This is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 *  USA.
 */
 
/* Uses portions of the VNC Viewer project by Lucas Gomez */

/*
 * x11Windows.cpp - Wrap commands to the Symbian window system, and provide stubs
 */

#include "x11Windows.h"
#include "x11parsers.h"
#include <sys\unistd.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <sys\shm.h>
#include <sys\sem.h>
#include "keysyms.h"

// Time out value for select, to detect connection breaks
const int KTimeOutSecs = 10;

const char* KTitleFormat = "%s";


/**
Access the window system

@return CSymbianWindowSystem* The window system
*/
CSymbianWindowSystem* WindowSystem()
	{
	// Test to see if the window system has been created
	CSymbianWindowSystem* ws = static_cast<CSymbianWindowSystem*>(Dll::Tls()); 
	if (!ws)
		{
		// Create the window system class
		ws = new CSymbianWindowSystem();
		if (ws)
			{
			// Initialise the system
			TRAPD(err, ws->ConstructL());
			if (KErrNone != err)
				{
				delete ws;
				ws = NULL;
				}
			else
				{
				// Put the window system into Thread local storage
				Dll::SetTls(ws);
				}
			}
		}
	return ws;
	}

/**
Destroy the window system
*/
void DestroyWindowSystem()
	{
	CSymbianWindowSystem* ws = WindowSystem();
	delete ws;
	Dll::SetTls(NULL);
	
	// Kill our process
	RProcess process;
	process.Kill(0);
	}

/**
Empty constructor
*/
CSymbianWindowSystem::CSymbianWindowSystem()
	{
	}

/**
2nd phase constructor
*/
void CSymbianWindowSystem::ConstructL()
	{	
	// Connect to the window and file servers
	User::LeaveIfError(iWsSession.Connect());	
	User::LeaveIfError(iFs.Connect());	
	
	// Create the option parse
	iOptionParser = COptionParser::NewL();

	// Create the IPC mechanism
	iVncIpc = CVncIpc::NewL(EFalse);
	
	// Create an active scheduler if one does not exist
	if (!CActiveScheduler::Current())
		{
		iAs = new(ELeave) CActiveScheduler();
		CActiveScheduler::Install(iAs);
		}

	// Create the current display parameters
	SetupDisplayL();		
	iDesktopWidget.widgetId = 0;
	}

/**
Destructor
*/
CSymbianWindowSystem::~CSymbianWindowSystem()
	{
	// Delete the IPC mechanism, palette and option parse
	delete iVncIpc;
	delete iPalette;
	delete iOptionParser;
	
	// Remove any memory
	while(iAllocatedMemory.Count() > 0)
		{
		User::Free(iAllocatedMemory[0]);
		iAllocatedMemory.Remove(0);
		}
	iAllocatedMemory.Close();

	// Destroy the allocated palettes
	iAllocatedPalettes.ResetAndDestroy();

	// Destroy the allocated bitmaps
	iBitmaps.ResetAndDestroy();
	
	// Destroy the inputs and events
	iAppInputs.ResetAndDestroy();
	iAppEvents.ResetAndDestroy();
	
	// Reset the actions
	iActions.Close();

	// Delete clipboard and bitmaps		
	delete iClipboardText;
	delete iJpegBitmap;
	delete iBitGc;
	delete iBitmapDevice;
	delete iMainBitmap;
	
	// Delete any allocated active scheduler
	delete iAs;

	// Close access to the servers
	iWsSession.Close();
	iFs.Close();
	}

/**
Allocate memory, and watch it
@param aSize Size of memory to allocates in bytes
@return TAny* The allocated memory
*/
TAny* CSymbianWindowSystem::AllocMemoryL(TInt aSize)
	{
	TAny* memory = User::AllocL(aSize);
	// Keep a track on the memory
	AddMemory(memory);	
	return memory;
	}

/**
Free memory, and stop watching it
@param aMemory Memory to free
*/
void CSymbianWindowSystem::FreeMemory(TAny* aMemory)
	{
	// Stop tracking the memory
	RemoveMemory(aMemory);
	User::Free(aMemory);
	}

/**
Add memory to the watcher
@param aMemory The memory to add
*/
void CSymbianWindowSystem::AddMemory(TAny* aMemory)
	{
	iAllocatedMemory.Append(aMemory);
	}
	
/**
Remove memory from the watcher
@param aMemory The memory to remove
*/
void CSymbianWindowSystem::RemoveMemory(TAny* aMemory)
	{
	TInt pos = iAllocatedMemory.Find(aMemory);
	if (KErrNotFound != pos)
		iAllocatedMemory.Remove(pos);
	}

/**
Setup internal display parameters for the application to query
*/
void CSymbianWindowSystem::SetupDisplayL()
	{	
	CWsScreenDevice* device = new(ELeave) CWsScreenDevice(iWsSession);
	CleanupStack::PushL(device);

	User::LeaveIfError(device->Construct());
		
	iDisplay.byte_order = LSBFirst;

	// Get stats about the screen
	TInt currScreenMode = device->CurrentScreenMode();
	TDisplayMode displayMode = device->GetScreenModeDisplayMode(currScreenMode);
	TPixelsAndRotation pixelsAndRotation;
	device->GetScreenModeSizeAndRotation(currScreenMode, pixelsAndRotation);
	
	// Set up the screen struct
	iScreen.display = &iDisplay;
	iScreen.width = pixelsAndRotation.iPixelSize.iWidth;
	iScreen.height = pixelsAndRotation.iPixelSize.iHeight;

	iVisual.visualid = (VisualID)displayMode;

	// Get the screen information
	TSymbDisplayInfo displayInfo;
	GetDisplayInfo(displayMode, displayInfo);
	iScreen.root_depth = displayInfo.iDepth;
	iVisual.c_class = displayInfo.iClass;
	iVisual.bits_per_rgb = displayInfo.iBpp;
	iVisual.red_mask = displayInfo.iRedMask;
	iVisual.green_mask = displayInfo.iGreenMask;
	iVisual.blue_mask = displayInfo.iBlueMask;
	iVisual.map_entries = 1 << iScreen.root_depth;
	
	iDisplay.nscreens = 1;
	iDisplay.screens = &iScreen;
	iScreen.root_visual = &iVisual;

	// Allocate a palette
	iPalette = CPalette::NewDefaultL(EColor256);
	const TColor256Util* defaultPalette = TColor256Util::Default();
	for(TInt count = 0; count < 256; ++count)
		iPalette->SetEntry(count, defaultPalette->Color256(count));

	iScreen.cmap = (Colormap)iPalette;
		
	CleanupStack::PopAndDestroy(device); // device
	}

/**
Initialize the Application parameters. This takes a structure of possible
command line parameters, and parses the command line passed to the process. 
It creates an internal structure to store details of the passed in options and 
removes them from the command line strings.
@param app_context_return The application context
@param options The list of possible command line options
@param argc_in_out The number of command line options passed to the program
@param argv_in_out The command line options passed to the program
@return Widget The desktop widge
*/
Widget CSymbianWindowSystem::InitAppL(XtAppContext* aApp_context_return,
    XrmOptionDescList	aOptions,
    Cardinal		aNum_options,
    int*		aArgc_in_out,
    String*		aArgv_in_out)
	{
	// Create the context
	aApp_context_return = new(ELeave) XtAppContext;
	CleanupStack::PushL(aApp_context_return);

	if (*aArgc_in_out > 1)
		{
		// Create an array to contain the strings
		CDesC8Array* argArray = new(ELeave) CDesC8ArrayFlat(*aArgc_in_out);
		CleanupStack::PushL(argArray);
		
		// Populate the array
		for(TInt count = 1; count < *aArgc_in_out; ++count)
			{
			TPtrC8 ptrArg((TUint8*)aArgv_in_out[count]);
			argArray->AppendL(ptrArg);
			}
	
		// Add the memory for the first parameter (which would not be freed)
		AddMemory(aArgv_in_out[0]);
		AddMemory(aArgv_in_out);
			
		// Free the old memory location (all the args have been stored in one cell)		
		User::Free(aArgv_in_out[1]);
	
		// Now parse the options
		TInt count = 0; 
		while(count < argArray->MdcaCount())
			{
			if (iOptionParser->ParseArgumentL(aNum_options, aOptions, argArray->MdcaPoint(count)))
				argArray->Delete(count);
			else
				count++;
			}
		
		// Now re-create the remaining arguments
		// Calculate the length
		TInt length = 0;
		for (TInt count = 0; count < argArray->MdcaCount(); ++count)
			length += argArray->MdcaPoint(count).Length() + 1; // Length of each arg + \0
		
		aArgv_in_out[1] = static_cast<char*>(AllocMemoryL(sizeof(char*) * length));
		Mem::FillZ(aArgv_in_out[1], length);
		
		TInt argNo = 1;
		char* pArg = aArgv_in_out[argNo];
		
		// Now store each string
		for(TInt count = 0; count < argArray->MdcaCount(); ++count)
			{
			TPtrC8 ptrArg(argArray->MdcaPoint(count));
			Mem::Copy(pArg, ptrArg.Ptr(), ptrArg.Length());
			aArgv_in_out[argNo++] = pArg;
			pArg += ptrArg.Length() + 1;		
			}
		
		*aArgc_in_out = argArray->MdcaCount() + 1;
		CleanupStack::PopAndDestroy(argArray); // argArray
		}

	CleanupStack::Pop(aApp_context_return); // context
	
	AddMemory(aApp_context_return);	
	return &iDesktopWidget;
	}

/**
Fills in data elements in a structure with data from the command line, according
to a resource structure.
@param aBase The base of the data structure
@param aResources The resource list
@param aNum_resources The number of resources in the resource list
*/
void CSymbianWindowSystem::GetAppResourcesL(XtPointer aBase, 
	    XtResourceList 	aResources,
	    Cardinal 		aNum_resources)
	{
	for(TUint count = 0; count < aNum_resources; ++count)
		{
		const XtResource& resource = aResources[count];
		TInt pos = resource.resource_offset;
		char* ptrBase = static_cast<char*>(aBase) + pos;
		Mem::Copy(ptrBase, (void*)&resource.default_addr, resource.resource_size);
		
		iOptionParser->UpdateResourceL(resource, ptrBase);
		}
	}

/**
Add widget actions to the applications resources. Used in X11 to know which function
to callback when on when an event from the UI occurs
@param aActions The list of actions
@param aNum_actions The number of actions in the action list
*/
void CSymbianWindowSystem::AppAddActionsL(XtActionList aActions,
	    Cardinal 		aNum_actions)
	{
	for(TUint count = 0; count < aNum_actions; ++count)
		{
		const XtActionsRec& action = aActions[count];
		iActions.Append(action);
		}
	}
	
/**
Get a list of display modes which match certain criteria.
@param aVinfo_mask The criteria on which to match
@param aVinfo_template The specific criteria to match
@param aNitems_return The number of matching display modes
@return XVisualInfo* A list of matching visual
*/
XVisualInfo* CSymbianWindowSystem::GetVisualInfoL(long aVinfo_mask,
    XVisualInfo* 	aVinfo_template,
   	int*			aNitems_return)
	{
	// Now we're looking at display modes
	// Allocate memory for each display mode (this is slightly wastful of memory but will
	// speed up operation)
	XVisualInfo* retInfo = (XVisualInfo*)AllocMemoryL(sizeof(XVisualInfo) * EColorLast);
	
	*aNitems_return = 0;

	XVisualInfo tempInfo;	
	Mem::FillZ(&tempInfo, sizeof(XVisualInfo));

	// Loop through each mode
	for (TInt mode = (TInt)ENone; mode < (TInt)EColorLast; ++mode)
		{
		TDisplayMode currMode = (TDisplayMode) mode;

		// Get the display information
		TSymbDisplayInfo displayInfo;
		GetDisplayInfo(currMode, displayInfo);

		// Check if the criteria points to interest
		TBool interested = ETrue;
		if (((aVinfo_mask & VisualDepthMask) && (displayInfo.iDepth != aVinfo_template->depth)) ||
		    ((aVinfo_mask & VisualClassMask) && (displayInfo.iClass != aVinfo_template->c_class)) || 
		    ((aVinfo_mask & VisualColormapSizeMask) && ((1 << displayInfo.iDepth) != aVinfo_template->colormap_size)))
			{
			interested = EFalse;
			}

		// Yes, we are interested in this display mode
		if (interested)
			{
			// Allocate a visual
			Visual* vis = (Visual*)AllocMemoryL(sizeof(Visual));
			vis->visualid = (VisualID)mode;
			vis->c_class = displayInfo.iClass;
			vis->red_mask = displayInfo.iRedMask;
			vis->green_mask = displayInfo.iGreenMask;
			vis->blue_mask = displayInfo.iBlueMask;
			vis->bits_per_rgb = displayInfo.iBpp;
			vis->map_entries = 0;

			tempInfo.visual = vis;			
			tempInfo.visualid = (VisualID)mode;
			tempInfo.depth = displayInfo.iDepth;
			tempInfo.colormap_size = 1 << displayInfo.iDepth;
			tempInfo.c_class = displayInfo.iClass;
			tempInfo.red_mask = displayInfo.iRedMask;
			tempInfo.green_mask = displayInfo.iGreenMask;
			tempInfo.blue_mask = displayInfo.iBlueMask;
			tempInfo.bits_per_rgb = displayInfo.iBpp;
			Mem::Copy(&retInfo[*aNitems_return], &tempInfo, sizeof(XVisualInfo));
			++(*aNitems_return);
			Mem::FillZ(&tempInfo, sizeof(XVisualInfo));
			}
		}

	// Free memory if we're not interested in any		
	if (*aNitems_return == 0)
		{
		FreeMemory(retInfo);
		retInfo = NULL;		
		}
	
	return retInfo;
	}

/**
Get a list of supported bitmap modes 
@param aCount_return The number of supported bitmap modes
@return XPixmapFormatValues* Information about the supported bitmaps
*/
XPixmapFormatValues* CSymbianWindowSystem::ListPixmapFormatsL(int* aCount_return)
	{
	// Now we're looking at pixmaps 
	// Allocate memory for each display mode (this is slightly wastful of memory but will
	// speed up operation)
	XPixmapFormatValues* retFormatValues = (XPixmapFormatValues*)AllocMemoryL(sizeof(XPixmapFormatValues) * EColorLast);

	*aCount_return = 0;
	XPixmapFormatValues* currFormatValue = retFormatValues;

	// Loop through the display modes
	for (TInt mode = (TInt)ENone; mode < (TInt)EColorLast; ++mode)
		{
		TDisplayMode currMode = (TDisplayMode) mode;

		TSymbDisplayInfo displayInfo;
		GetDisplayInfo(currMode, displayInfo);

		if (displayInfo.iDepth > 0)
			{
			currFormatValue->depth = displayInfo.iDepth;
			currFormatValue->bits_per_pixel = displayInfo.iBpp;
			currFormatValue->scanline_pad = displayInfo.iScanLinePad;			
			++currFormatValue;
			++(*aCount_return);
			}
		}
		
	return retFormatValues;
	}
	
/**
Get display information about a display mode
@param aDisplayMode The display mode
@param aDisplayInfo Display information about the display mode
*/
void CSymbianWindowSystem::GetDisplayInfo(TDisplayMode aDisplayMode, TSymbDisplayInfo& aDisplayInfo)	
	{
	Mem::FillZ(&aDisplayInfo, sizeof(TSymbDisplayInfo));
	
	switch(aDisplayMode)	
		{
		case ENone:
		case ERgb:
			break;
		case EGray2:
				{
				aDisplayInfo.iDepth = 2;
				aDisplayInfo.iClass = TrueColor;
				}
			break;
		case EGray4:
		case EGray16:
		case EGray256:
			// Ignore other grayscales
			break;
		case EColor16: 
				{
				aDisplayInfo.iDepth = 4;
				aDisplayInfo.iClass = TrueColor;
				}
			break;
		case EColor256:
				{
				aDisplayInfo.iDepth = 8;
				aDisplayInfo.iClass = PseudoColor;
				}
			break;
		case EColor64K:
				{
				aDisplayInfo.iDepth = 16;
				aDisplayInfo.iClass = TrueColor;
				aDisplayInfo.iBlueMask = 0x1F;
				aDisplayInfo.iGreenMask = 0x7E0;
				aDisplayInfo.iRedMask = 0xF800;
				}
			break;
		case EColor4K:
				{
				aDisplayInfo.iDepth = 12;
				aDisplayInfo.iClass = TrueColor;
				aDisplayInfo.iBlueMask = 0x0F;
				aDisplayInfo.iGreenMask = 0xF0;
				aDisplayInfo.iRedMask = 0x0F00;
				}
			break;
		case EColor16M:
				{
				aDisplayInfo.iDepth = 24;
				aDisplayInfo.iClass = TrueColor;
				aDisplayInfo.iBlueMask = 0x000000FF;
				aDisplayInfo.iGreenMask = 0x0000FF00;
				aDisplayInfo.iRedMask = 0x00FF0000;
				}
			break;
		case EColor16MU:
		case EColor16MA:
				{
				aDisplayInfo.iDepth = 32;
				aDisplayInfo.iClass = TrueColor;
				aDisplayInfo.iBlueMask = 0x000000FF;
				aDisplayInfo.iGreenMask = 0x0000FF00;
				aDisplayInfo.iRedMask = 0x00FF0000;
				}
			break;
		default:
			break;
		}

	aDisplayInfo.iBpp = aDisplayInfo.iDepth;
	}

/**
Get system values
@param aWidget Unused
@param aVaList A list of values with pointers following in which to copy memory
*/
void CSymbianWindowSystem::GetValuesL(Widget /* aWidget */, VA_LIST aVaList)
	{
	while(1)
		{
		// Get the next item
		TUint8* name = VA_ARG(aVaList, TUint8*);
		if (!name)
			break;
		
		TPtrC8 ptrName(name);

		// Title of the window
		if (ptrName.Compare(TPtrC8((TUint8*)XtNtitle)) == 0)
			{
			// The parameter is a char*
			char** format = VA_ARG(aVaList, char**);
			*format = (char*)KTitleFormat;
			}
			
		// Position and size of windows
		if (ptrName.Compare(TPtrC8((TUint8*)XtNgeometry)) == 0)
			{
			// The parameter is a char*
			char** geometry = VA_ARG(aVaList, char**);
			
			iGeometry.Format(_L8("0x0+%d+%d"), iScreen.width, iScreen.height);
			iGeometry.ZeroTerminate();
			*geometry = (char*)(iGeometry.Ptr());
			}
		}	
	}
	
/**
Create a system colormap for the display mode
@param aVisual The current visual (display mode)
@param aAlloc Unused
@return Colormap The colormap or rather a point to a palette
*/
Colormap CSymbianWindowSystem::CreateColormapL(Visual* aVisual, int /* aAlloc */)
	{
	Colormap colormap = 0;
	
	// Ignore the alloc parameter, allocate a new palette for the display mode
	CPalette* palette = CPalette::NewDefaultL((TDisplayMode)aVisual->visualid);
	iAllocatedPalettes.Append(palette);	

	// Nasty, but pass the palette back as a colormap
	colormap = (Colormap)palette;
	return colormap;
	}
		
/**
Query the colors in a colormap
@param aColormap The colormap or palette
@param aDefs_in_out The colors to fill in 
@param aNcolors The number of colors in the structure
*/
void CSymbianWindowSystem::QueryColors(Colormap aColormap, XColor* aDefs_in_out, int aNcolors)
	{
	CPalette* palette = (CPalette*)aColormap;
	
	// For each color fill in the color
	for(TInt count = 0; count < aNcolors; ++count)
		{
		XColor color = aDefs_in_out[count];
		color.pixel = count;
		
		// Get the Rgb value from the palette
		TRgb rgb = palette->GetEntry(count);
		
		// All X colors are on a scale from 0 to 65535, so we need to scale the rgb values from 0 to 255
		// We also want to make sure that 0 (black) is the same on both as is 255/65535 (white)
		color.red = SymbToXColor(rgb.Red());
		color.green = SymbToXColor(rgb.Green());
		color.blue = SymbToXColor(rgb.Blue());
		color.flags = DoRed | DoGreen | DoBlue;
		}
	}
	
/**
Allocate a color. Actually the colors are already allocated, this finds the nearest color
to the one desired
@param aColormap The colormap or palette
@param aScreen_in_out The color desired, and nearest one found
*/
void CSymbianWindowSystem::AllocColor(Colormap aColormap, XColor* aScreen_in_out)
	{
	// This usually allocates a color in the palette, but in Symbian it is already allocated,
	// We just find the closest color match
	CPalette* palette = (CPalette*)aColormap;

	TRgb existingColor(0);
	existingColor.SetRed(XToSymbColor(aScreen_in_out->red));
	existingColor.SetGreen(XToSymbColor(aScreen_in_out->green));
	existingColor.SetBlue(XToSymbColor(aScreen_in_out->blue));
	
	// Now find the nearest colour and index
	TRgb nearestColor = palette->NearestEntry(existingColor);
	TInt nearestIndex = palette->NearestIndex(existingColor);
	
	// Store the new values
	aScreen_in_out->pixel = nearestIndex;
	aScreen_in_out->red = SymbToXColor(nearestColor.Red());
	aScreen_in_out->green = SymbToXColor(nearestColor.Green());
	aScreen_in_out->blue = SymbToXColor(nearestColor.Blue());
	}
		
/**
Create the main bitmap to store the screen
@param aVisual The visual or display mode
@param aDepth The depth of the image in bits
@param aWidth The width of the image in pixels
@param aHeight The height of the image in pixels
@param aBytes_per_line Unused
@return XImage* The XImage style image
*/
XImage* CSymbianWindowSystem::CreateImageL(Visual* aVisual, int aDepth, int aWidth, int aHeight, int /* aBytes_per_line */)
	{
	// Create the bitmap
	iMainBitmap = new(ELeave) CFbsBitmap();
	User::LeaveIfError(iMainBitmap->Create(TSize(aWidth, aHeight), (TDisplayMode)aVisual->visualid));
	
	// Create the device and GC to manipulate the image
    iBitmapDevice = CFbsBitmapDevice::NewL(iMainBitmap);
	iBitGc = CFbsBitGc::NewL();
	iBitGc->Activate(iBitmapDevice);
	
	// Create the X Windows Bitmap
	XImage* newImage = (XImage*)AllocMemoryL(sizeof(XImage));
	Mem::FillZ(newImage, sizeof(XImage));
	newImage->depth = aDepth;
	newImage->width = aWidth;
	newImage->height = aHeight;

	// Save the display mode
	iDisplayMode = (TDisplayMode)aVisual->visualid;

	// Now send the duplicate request to the UI
	struct vncmsg msg;
	msg.iData.iCreateBitmap.iHandle = iMainBitmap->Handle();
	iVncIpc->SendMessageL(ECreateBitmap, msg);

	TInt bytesPerPixel = TDisplayModeUtils::NumDisplayModeBitsPerPixel((TDisplayMode)aVisual->visualid) / 8;
	newImage->bytes_per_line = aWidth * bytesPerPixel;
		
	// Set the memory address
	iMainBitmap->LockHeap();
	newImage->data = (char*)iMainBitmap->DataAddress();
	iMainBitmap->UnlockHeap();
	
	return newImage;
	}

/**
Get the password from the user
@param aPrompt Unused
@return char* The password
*/
char* CSymbianWindowSystem::GetPassL(const char* /* aPrompt */)
	{
	// Send the password request to the UI
	struct vncmsg msg;
	iVncIpc->SendMessageL(EShowPasswordReq, msg);
		
	// Now get the response
	iVncIpc->RecvMessageL(msg);

	if (msg.iMsgType == EShowPasswordResp)
		{
		char* password = (char*)AllocMemoryL(strlen(msg.iData.iPassword) + 1);
		strcpy(password, msg.iData.iPassword);
		return password;
		}
	else
		User::Leave(KErrArgument);
		
	return NULL;
	}

/**
Add an input to the process routine
@param aSource The descriptor
@param aCondition Unused
@param aProc The function to call if the descriptor is signalled
@param aClosure Unused
@return XtInputId An id for the input
*/
XtInputId CSymbianWindowSystem::AppAddInputL(int aSource,
    XtPointer 		/* aCondition */,
	XtInputCallbackProc aProc,
	XtPointer 		/* aClosure */)
	{
	TAppInput* input = new(ELeave) TAppInput();
	input->iSource = aSource;
	input->iProc = aProc;
	iAppInputs.Append(input);
	return input;
	}
	
/**
Remove an input from the process routine
@param aInput The input to remove
*/
void CSymbianWindowSystem::AppRemoveInputL(XtInputId aInput)
	{
	TInt pos = iAppInputs.Find((TAppInput*)aInput);
	if (KErrNotFound != pos)
		{
		TAppInput* input = iAppInputs[pos];
		iAppInputs.Remove(pos);
		delete input;	
		}
	}
	
/**
The process function, this processes registered inputs an UI events
*/
void CSymbianWindowSystem::AppProcessEventL()
	{
	// Create the set
	int maxfd = 0;
	fd_set readSet;
	FD_ZERO(&readSet);

	// Add all the inputs
	for(TInt count = 0; count < iAppInputs.Count(); ++count)
		{
		const TAppInput* input = iAppInputs[count];
		FD_SET(input->iSource, &readSet);
		if (input->iSource > maxfd)
			maxfd = input->iSource;
		}
	
	// Now add the event fifo
	FD_SET(iVncIpc->EventFifo(), &readSet);
	if (iVncIpc->EventFifo() > maxfd)
		maxfd = iVncIpc->EventFifo();
	
	// Run the select, would use a timeout but there is a defect
	int ret = select(maxfd + 1, &readSet, NULL, NULL, NULL);
	if (ret > 0)
		{
		// Check to see if the event fifo has trigger
		if (FD_ISSET(iVncIpc->EventFifo(), &readSet))
			{
			// Process the message
			TRAPD(err, ProcessIncomingMessageL());
			if (KErrNone != err)
				{
				fprintf(stderr, "Unknown message");
				}
			return;
			}
		else
			{
			// Check to see which input triggered
			for(TInt count = 0; count < iAppInputs.Count(); ++count)
				{
				TAppInput* input = iAppInputs[count];
				// Check to see if the event triggered
				if (FD_ISSET(input->iSource, &readSet))
					{
					// If there is an excpetion, let the read or write catch it
					void* inputV = (void*)input;
					
					// Call the callback function
					input->iProc(NULL, &input->iSource, &inputV);
					return;
					}
				}
			}
		}
		
	// We should not have got here, but we have, loop but don't take the processor time
	User::After(1000);
	}

/**
Process an incoming UI message
*/
void CSymbianWindowSystem::ProcessIncomingMessageL()
	{
	struct vncmsg msg;
	iVncIpc->RecvEventL(msg);
	switch(msg.iMsgType)
		{
		case EGraphicsExposeEvent:
			ProcessGraphicsExposeEvent(msg);
			break;
		case EMouseEvent:
			ProcessMouseEventL(msg);
			break;
		case EKeyEvent:
			ProcessKeyEventL(msg);
			break;
		case ECopyClipboardToVNCReq:
			ProcessCopyClipboardToServerL(msg);
			break;
		case ECopyClipboardFromVNCReq:
			ProcessCopyClipboardFromServerL(msg);
			break;
		case ERefreshReq:
			ProcessRefreshL(msg);
			break;
		case ESetFullscreen:
			ProcessForegroundL(msg);
			break;
		case EQuitReq:
			ProcessQuitL(msg);
			break;
		}
	}

/**
Process a graphics expose event
@param aMsg The event
*/
void CSymbianWindowSystem::ProcessGraphicsExposeEvent(struct vncmsg& aMsg)
	{
	// Find any FocusChangeEvent's 
	for (TInt count = 0; count < iAppEvents.Count(); ++count)
		{
		const TAppEvent* event = iAppEvents[count];
		if (event->iEventMask == FocusChangeMask)
			{
			XEvent xevent;
			xevent.type = GraphicsExpose;
			xevent.xexpose.x = aMsg.iData.iGraphicsExposeEvent.iX;
			xevent.xexpose.y = aMsg.iData.iGraphicsExposeEvent.iY;
			xevent.xexpose.width = aMsg.iData.iGraphicsExposeEvent.iWidth;
			xevent.xexpose.height = aMsg.iData.iGraphicsExposeEvent.iHeight;
			Boolean passOn = True;
			event->iProc(0, event->iClosure, &xevent, &passOn);
			if (passOn == False)
				return;
			}
		}
	}

/**
Process a mouse event
@param aMsg The event
*/
void CSymbianWindowSystem::ProcessMouseEventL(struct vncmsg& aMsg)
	{
	_LIT8(KSendRFBEvent, "SendRFBEvent");
	const XtActionsRec& action = FindActionL(KSendRFBEvent);

	XEvent xevent;
	xevent.xbutton.x = aMsg.iData.iMouseEvent.iX;
	xevent.xbutton.y = aMsg.iData.iMouseEvent.iY;

	switch(aMsg.iData.iMouseEvent.iType)
		{
		case EMouse1Down:
			xevent.type = ButtonPress;
			xevent.xbutton.button = Button1;
			xevent.xbutton.state = Button1MotionMask;
			break;	
		case EMouse1Up:
			xevent.type = ButtonRelease;
			xevent.xbutton.button = Button1;
			xevent.xbutton.state = Button1MotionMask;
			break;	
		case EMouse2Down:
			xevent.type = ButtonPress;
			xevent.xbutton.button = Button2;
			xevent.xbutton.state = Button2MotionMask;
			break;	
		case EMouse2Up:
			xevent.type = ButtonRelease;
			xevent.xbutton.button = Button2;
			xevent.xbutton.state = Button2MotionMask;
			break;	
		case EMouse3Down:
			xevent.type = ButtonPress;
			xevent.xbutton.button = Button3;
			xevent.xbutton.state = Button3MotionMask;
			break;	
		case EMouse3Up:
			xevent.type = ButtonRelease;
			xevent.xbutton.button = Button3;
			xevent.xbutton.state = Button3MotionMask;
			break;	
		case EMouseDrag:
			xevent.type = MotionNotify;
			xevent.xmotion.state = Button1MotionMask;
			break;	
		case EMouseMove:
			xevent.type = MotionNotify;
			xevent.xmotion.state = 0;
			break;	
		}
	
	Cardinal num_params = 0;
	action.proc(0, &xevent, NULL, &num_params);
	}

/**
Process a key expose event
@param aMsg The event
*/
void CSymbianWindowSystem::ProcessKeyEventL(struct vncmsg& aMsg)
	{
	_LIT8(KSendRFBEvent, "SendRFBEvent");
	const XtActionsRec& action = FindActionL(KSendRFBEvent);

	XEvent xevent;
	char* params[2];
	char eventtype[10];
	char keycode[10];
	params[0] = eventtype;
	params[1] = keycode;
	switch(aMsg.iData.iKeyEvent.iType)
		{
		case EKeyDown:
			strcpy(params[0], "keydown");
			break;							
		case EKeyUp:
			strcpy(params[0], "keyup");
			break;							
		case EKey:
			strcpy(params[0], "key");
			break;							
		}
	
	strcpy(params[1], aMsg.iData.iKeyEvent.iKeyChar);	
	
	Cardinal num_params = 2;
	action.proc(0, &xevent, params, &num_params);
	}
	
/**
Process a copy clipboard to server event
@param aMsg The event
*/
void CSymbianWindowSystem::ProcessCopyClipboardToServerL(struct vncmsg& /* aMsg */)
	{
	// The UI is wanting access to synchronise with us
	struct vncmsg msg;
	iVncIpc->SendMessageL(ECopyClipboardToVNCAck, msg);
	iVncIpc->RecvMessageL(msg);

	iClipboardLength = msg.iData.iCopyClipboard.iLength;

	// Now get the clipboard text
	delete [] iClipboardText;
	iClipboardText = NULL;
	iClipboardText = new(ELeave) char[iClipboardLength + 1];
	Mem::FillZ(iClipboardText, iClipboardLength + 1);
	iVncIpc->RecvExtraData(iClipboardText, iClipboardLength);

	_LIT8(KSelectionToVNC, "SelectionToVNC");
	const XtActionsRec& action = FindActionL(KSelectionToVNC);

	XEvent xevent;
	xevent.type = SelectionRequest;
	xevent.xselectionrequest.time = time(NULL);

	char selectiontype[10];
	char* params = selectiontype;
	strcpy(selectiontype, "always");
	Cardinal num_params = 1;
	action.proc(0, &xevent, &params, &num_params);
	}

/**
Process a copy clipboard from server event
@param aMsg The event
*/
void CSymbianWindowSystem::ProcessCopyClipboardFromServerL(struct vncmsg& /* aMsg */)
	{
	_LIT8(KSelectionFromVNC, "SelectionFromVNC");
	const XtActionsRec& action = FindActionL(KSelectionFromVNC);

	XEvent xevent;
	xevent.type = SelectionRequest;
	xevent.xselectionrequest.time = time(NULL);

	char selectiontype[10];
	char* params = selectiontype;
	strcpy(selectiontype, "always");
	Cardinal num_params = 1;
	action.proc(0, &xevent, &params, &num_params);
	}

/**
Process a refresh event
@param aMsg The event
*/
void CSymbianWindowSystem::ProcessRefreshL(struct vncmsg& /* aMsg */)
	{
	_LIT8(KSendRFBEvent, "SendRFBEvent");
	const XtActionsRec& action = FindActionL(KSendRFBEvent);

	XEvent xevent;

	char update[10];
	char* params = update;
	strcpy(update, "fbupdate");
	Cardinal num_params = 1;
	action.proc(0, &xevent, &params, &num_params);
	}

/**
Process a foreground event
@param aMsg The event
*/
void CSymbianWindowSystem::ProcessForegroundL(struct vncmsg& aMsg)
	{
	_LIT8(KToggleFullscreen, "ToggleFullScreen");
	const XtActionsRec& action = FindActionL(KToggleFullscreen);

	XEvent xevent;

	char update[10];
	char* params = update;
	if (aMsg.iData.iSetFullscreen.iFullscreen)
		strcpy(update, "on");
	else
		strcpy(update, "off");
	
	Cardinal num_params = 1;
	action.proc(0, &xevent, &params, &num_params);
	}

/**
Process a quit event
@param aMsg The event
*/
void CSymbianWindowSystem::ProcessQuitL(struct vncmsg& /* aMsg */)
	{
	iVncIpc->Quit();	

	_LIT8(KQuit, "Quit");
	const XtActionsRec& action = FindActionL(KQuit);
			
	XEvent xevent;
	Cardinal num_params = 0;
	action.proc(0, &xevent, NULL, &num_params);	
	}
	
/**
Find an action in the action list
@param aAction The action name
@return const XtActionsRec& The action record
*/
const XtActionsRec& CSymbianWindowSystem::FindActionL(const TDesC8& aAction)
	{
	// Find an action
	for (TInt count = 0; count < iActions.Count(); ++count)
		{
		const XtActionsRec& action = iActions[count];
		TPtrC8 ptrAction((TUint8*)action.string);
		if (ptrAction.CompareF(aAction) == 0)
			{				
			return action;	
			}
		}				
	User::Leave(KErrNotFound);
	return XtActionsRec();
	}

/**
Add an event handler, such as GraphicsExpose
@param aEvent_mask The event mask
@param aNonmaskable Unused
@param aProc The callback function
@param aClosure The closure function
*/
void CSymbianWindowSystem::AddEventHandlerL(EventMask aEvent_mask,
    _XtBoolean /* aNonmaskable */,
    XtEventHandler aProc,
	XtPointer aClosure)
	{
	TAppEvent* event = new(ELeave) TAppEvent();
	event->iEventMask = aEvent_mask;
	event->iProc = aProc;
	event->iClosure = aClosure;
	iAppEvents.Append(event);
	}

/**
Prepare a bitmap memory write by locking the bitmap
*/
void CSymbianWindowSystem::PrepareMemoryTransfer()
	{
	// Lock the bitmap
	if (iMainBitmap)
		{
		iMainBitmap->LockHeap();
		}
	}

/**
End the bitmap memory write by unlocking the bitmap
*/
void CSymbianWindowSystem::EndMemoryTransfer()
	{
	// Unlock the bitmap
	if (iMainBitmap)
		{
		iMainBitmap->UnlockHeap();
		}
	}

/**
Get the bitmap memory buffer
@return char* Pointer to the bitmap
*/
char* CSymbianWindowSystem::MemoryBuffer()
	{
	return (char*)iMainBitmap->DataAddress();
	}

/**
Put the image on the screen
@param aX The x position
@param aY The y position
@param aWidth The width of the image
@param aHeight The height of the image
*/
void CSymbianWindowSystem::PutImageL(int aX, int aY, int aWidth, int aHeight)
	{
	struct vncmsg msg;
	msg.iData.iPutImage.iX = aX;
	msg.iData.iPutImage.iY = aY;
	msg.iData.iPutImage.iWidth = aWidth;
	msg.iData.iPutImage.iHeight = aHeight;
	iVncIpc->SendMessageL(EPutImage, msg);
	}

/**
Change the GC (foreground color)
@param aMask The GC mask
@param aValues The values to change
*/
void CSymbianWindowSystem::ChangeGcL(int aMask, XGCValues* aValues)
	{
	if (aMask == GCForeground)
		{
		if (iDisplayMode == EColor256)
			{
			// We need to map the foreground to the nearest colour
			iForegroundColor = iPalette->GetEntry(aValues->foreground);
			}
		else
			{
			TSymbDisplayInfo displayInfo;
			GetDisplayInfo(iDisplayMode, displayInfo);
			
			TInt redShift = ffs(displayInfo.iRedMask) - 1;
			TInt greenShift = ffs(displayInfo.iGreenMask) - 1;
			TInt blueShift = ffs(displayInfo.iBlueMask) - 1;
			TInt redMax = displayInfo.iRedMask >> redShift;
			TInt greenMax = displayInfo.iGreenMask >> greenShift;
			TInt blueMax = displayInfo.iBlueMask >> blueShift;
			
			iForegroundColor.SetRed(((aValues->foreground & displayInfo.iRedMask) >> redShift) * 255 / redMax);
			iForegroundColor.SetGreen(((aValues->foreground & displayInfo.iGreenMask) >> greenShift) * 255 / greenMax);
			iForegroundColor.SetBlue(((aValues->foreground & displayInfo.iBlueMask) >> blueShift) * 255 / blueMax);
			}
		}
	}

/**
Fill a rectange on the main bitmap with the foreground colour
@param aX The x position
@param aY The y position
@param aWidth The width of the image
@param aHeight The height of the image
*/
void CSymbianWindowSystem::FillRectangleL(int aX, int aY, int aWidth, int aHeight)
	{
	if (iBitGc)
		{
		// Set the GC parameters
		iBitGc->SetBrushStyle(CGraphicsContext::ESolidBrush);
		iBitGc->SetBrushColor(iForegroundColor);
		iBitGc->SetPenColor(iForegroundColor);
		iBitGc->SetDrawMode(CGraphicsContext::EDrawModePEN);

		TRect rect(TPoint(aX, aY), TSize(aWidth, aHeight));
		iBitGc->DrawRect(rect);
		}
	}
	
/**
Convert JPEG data into a bitmap
@param aVisual The visual (display mode)
@param aWidth The width of the image
@param aHeight The height of the image
@param aData The jpeg data
@param aLength The length of the data
*/
void CSymbianWindowSystem::JpegConvertL(Visual* aVisual, int aWidth, int aHeight, char* aData, long aLength)
	{
	// Delete the existing Jpeg bitmap
	delete iJpegBitmap;
	iJpegBitmap = NULL;

	// Create a new jpeg
	iJpegBitmap = new(ELeave) CFbsBitmap();
	User::LeaveIfError(iJpegBitmap->Create(TSize(aWidth, aHeight), (TDisplayMode)aVisual->visualid));

	// Get the data stride	
	TInt bytesPerPixel = TDisplayModeUtils::NumDisplayModeBitsPerPixel((TDisplayMode)aVisual->visualid) / 8;
	iDataStride = aWidth * bytesPerPixel;

	// Convert the data
	TPtrC8 ptrBuf((TUint8*)aData, aLength);
	CJpegConverter::ConvertL(iFs, ptrBuf, iJpegBitmap);
	}

/**
Get the JPEG data
@return char* The data
*/
char* CSymbianWindowSystem::JpegGetDataL()
	{
	TSize size = iJpegBitmap->SizeInPixels();

	iJpegData = (char*)AllocMemoryL(iDataStride * size.iHeight);
	iJpegBitmap->LockHeap();
	Mem::Copy(iJpegData, iJpegBitmap->DataAddress(), iDataStride * size.iHeight);
	iJpegBitmap->UnlockHeap();
		
	return iJpegData;		
	}

/**
Discard the JPEG data
*/
void CSymbianWindowSystem::JpegDiscardData()
	{
	FreeMemory(iJpegData);
	iJpegData = NULL;
	delete iJpegBitmap;
	iJpegBitmap = NULL;
	}
	
/**
Get the clipboard data
@param aWidget The widget
@param aSelection The selection 
@param aTarget Unused
@param aCallback The callback function
@param aClosure The closure function
@param aXtTime The time at which the clipboard changed
*/
void CSymbianWindowSystem::GetSelectionValue(
    	Widget 		aWidget,
    	Atom 		aSelection,
    	Atom 		/* aTarget */,
	    XtSelectionCallbackProc aCallback,
   	 	XtPointer 		aClosure,
    	XtTime 		/* aXttime */)
	{
	if (iClipboardLength == 0)
		return;
		
	Atom type;
	int format;
	unsigned long length = iClipboardLength;
	aCallback(aWidget, aClosure, &aSelection, &type, iClipboardText, &length, &format);
	}
	
/**
Store the clipboard data
@param aBytes The clipboard data
@param aNBytes The length of clipboard data
*/
void CSymbianWindowSystem::StoreBytesL(const char* aBytes, int aNbytes)
	{
	// Calculate how much we can copy
	struct vncmsg msg;
	msg.iData.iCopyClipboard.iLength = aNbytes;
	iVncIpc->SendMessageL(ECopyClipboardTransfer, msg, (char*)aBytes, aNbytes);
	}
	
/**
Create a pixmap (bitmap)
@param aWidth The width of the bitmap
@param aHeight The height of the bitmap
@param aDepth The depth of the bitmap
*/
Pixmap CSymbianWindowSystem::CreatePixmapL(unsigned int aWidth,
	    unsigned int	aHeight,
    	unsigned int	aDepth)
	{
	CFbsBitmap* bitmap = new(ELeave) CFbsBitmap;
	iBitmaps.Append(bitmap);
	
	TDisplayMode mode;
	switch(aDepth)
		{
		case 32:
			mode = EColor16MU; 
			break;		
		case 16:
			mode = EColor64K; 
			break;		
		case 8:
			mode = EColor256; 
			break;	
		case 1:
			mode = EGray2; 
			break;	
		}
	
	User::LeaveIfError(bitmap->Create(TSize(aWidth, aHeight), mode));
	
	return (Pixmap)bitmap;
	}
	
/**
Copy an area from one bitmap to another
@param aSrc The source bitmap
@param aDest The destination bitmap
@param aSrc_x The source bitmap x position
@param aSrc_y The source bitmap y position
@param aWidth The width of the bitmap
@param aHeight The height of the bitmap
@param aDest_x The destination bitmap x position
@param aDest_y The destination bitmap y position
*/
void CSymbianWindowSystem::CopyAreaL(
	    Drawable		aSrc,
	    Drawable		aDest,
	    int				aSrc_x,
	    int				aSrc_y,
	    unsigned int	aWidth,
	    unsigned int	aHeight,
	    int				aDest_x,
	    int				aDest_y)
	{
	TPoint point(aDest_x, aDest_y);
	TRect rect(TPoint(aSrc_x, aSrc_y), TSize(aWidth, aHeight));
	
	// Copying from the pixmap
	CFbsBitmap* srcBitmap = reinterpret_cast<CFbsBitmap*>(aSrc);
	CFbsBitmap* destBitmap = reinterpret_cast<CFbsBitmap*>(aDest);
	
	if (iBitmaps.Find(srcBitmap) != KErrNotFound)
		{
		if (iBitGc)
			iBitGc->BitBlt(point, srcBitmap, rect);
		}
	else if (iBitmaps.Find(destBitmap) != KErrNotFound)
		{
		CFbsBitmapDevice* bitDevice = CFbsBitmapDevice::NewL(destBitmap);
		CleanupStack::PushL(bitDevice);
		CFbsBitGc* bitGc = CFbsBitGc::NewL();
		CleanupStack::PushL(bitGc);
		bitGc->Activate(bitDevice);
		bitGc->BitBlt(point, iMainBitmap, rect);
		CleanupStack::PopAndDestroy(2); // bitGc, bitDevice
		}
	}
	
/**
Free a pixmap (or bitmap)
@param aPixmap The pixmap to free
*/
void CSymbianWindowSystem::FreePixmapL(Pixmap aPixmap)
	{
	CFbsBitmap* bitmap = reinterpret_cast<CFbsBitmap*>(aPixmap);

	TInt pos = iBitmaps.Find(bitmap);
	if (pos != KErrNotFound)
		{
		iBitmaps.Remove(pos);
		delete bitmap;
		}
	}

/**
Create a bitmap with predetermined data
@param aWidth The width of the bitmap
@param aHeight The height of the bitmap
@param aDepth The depth of the bitmap
@param aData The data
@return Pixmap The new pixmap/bitmap
*/
Pixmap CSymbianWindowSystem::CreateBitmapWithDataL(int width, int height, int depth, unsigned char* data)
	{
	TInt bytesRequired = (width * height * depth) / 8;
	
	Pixmap pixmap = CreatePixmapL(width, height, depth);
	
	// Convert back to bitmap
	CFbsBitmap* bitmap = reinterpret_cast<CFbsBitmap*>(pixmap);
	
	// Lock the bitmap
	bitmap->LockHeap();
	
	TUint8* bitmapAddress = (TUint8*)bitmap->DataAddress();

	if (depth == 1)	
		{
		// We need to convert down to bit fields, and gray2 is back-to-front with bits
		TUint8* destBytes = bitmapAddress;
		TUint8* srcBytes = (TUint8*)data;
		TInt shift = 0;
		TUint8 currByte = 0;
		for(TInt y = 0; y < height; ++y)
			{
			for(TInt x = 0; x < width; ++x)
				{
				if (*(srcBytes++))
					{
					currByte |= 1 << shift;
					}
					
				if (++shift > 7)
					{
					shift = 0;
					*(destBytes++) = currByte;
					currByte = 0;
					}
				}
				
			if (width%8)
				{
				*(destBytes++) = currByte;					
				}
			shift = 0;
			currByte = 0;
			}
		}
	else
		Mem::Copy(bitmapAddress, data, bytesRequired);
	
	// Lock the bitmap
	bitmap->UnlockHeap();
		
	return pixmap;
	}

/**
Draw a bitmap onto the main bitmap
@param aPixmap The pixmap/bitmap to blit
@param aMask The pixmap/bitmap mask to blit
@param aX The x position on the destination bitmap
@param aY The y position on the destination bitmap
@param aWidth The width of the bitmap
@param aHeight The height of the bitmap
@param aHotpointX The hotpoint X position on the bitmap
@param aHotpointY The hotpoint Y position on the bitmap
*/
void CSymbianWindowSystem::DrawBitmapL(Pixmap aPixmap, Pixmap aMask, int aX, int aY, int aWidth, int aHeight, int aHotpointX, int aHotpointY)
	{
	// Copying from the pixmap
	CFbsBitmap* bitmap = reinterpret_cast<CFbsBitmap*>(aPixmap);
	CFbsBitmap* bitmapMask = reinterpret_cast<CFbsBitmap*>(aMask);

	if ((iBitmaps.Find(bitmap) != KErrNotFound) && (iBitmaps.Find(bitmapMask) != KErrNotFound))
		{
		TSize size_rect(aWidth, aHeight);
		TRect srcRect(TPoint(0, 0), size_rect);
		TPoint destPoint(TPoint(aX, aY));

		if (iBitGc)
			{
			iBitGc->SetBrushStyle(CGraphicsContext::ENullBrush);
			iBitGc->BitBltMasked(destPoint, bitmap, srcRect, bitmapMask, EFalse);
			}
		}

	struct vncmsg msg;
	msg.iData.iMoveSoftCursor.iX = aX + aHotpointX;
	msg.iData.iMoveSoftCursor.iY = aY + aHotpointY;
	iVncIpc->SendMessageL(EMoveSoftCursor, msg);
	}

/**
Create a bitmap with 1 bit data
@param aWidth The width of the bitmap
@param aHeight The height of the bitmap
@param aData The data
*/
Pixmap CSymbianWindowSystem::CreateBitmapFromDataL(int aWidth, int aHeight, const char* aData)
	{
	// Check the number of bytes required
	int bytesPerRow = (aWidth + 7) / 8;

	Pixmap pixmap = CreatePixmapL(aWidth, aHeight, 1);
	
	// Convert back to bitmap
	CFbsBitmap* bitmap = reinterpret_cast<CFbsBitmap*>(pixmap);
	
	TBitmapUtil bitutil = TBitmapUtil(bitmap);
	bitutil.Begin(TPoint(0, 0));

	int i, j, x;
	for (i = 0; i < aHeight; i++) {
		bitutil.SetPos(TPoint(0, i));
		x = 0;
		
		for (j = 0; j < bytesPerRow; j++) {
			int mask = 0x80;
			
			while (mask > 0) {
				if (x < aWidth) {
					if (aData[i * bytesPerRow + j] & mask) {
						bitutil.SetPixel(1);
					} else {
						bitutil.SetPixel(0);
					}
					bitutil.IncXPos();
				}
				x++;
				mask = mask >> 1;
			}
		}
	}

	bitutil.End();
	
	return pixmap;
	}

/**
Create a pixmap cursor from existing pixmaps
@param aSource The source pixmap
@param aMask The pixmap mask
@param aHotpointX The hotpoint x position on the destination bitmap
@param aHotpointY The hotpoint y position on the destination bitmap
@return Cursor The cursor
*/
Cursor CSymbianWindowSystem::CreatePixmapCursorL(Pixmap aSource,
	    Pixmap		aMask,
		unsigned int	aHotpointX,
    	unsigned int	aHotpointY)
	{
	// Copying from the pixmap
	CFbsBitmap* bitmap = reinterpret_cast<CFbsBitmap*>(aSource);
	CFbsBitmap* bitmapMask = reinterpret_cast<CFbsBitmap*>(aMask);
	
	struct vncmsg msg;
	msg.iData.iCreatePixmapCursorReq.iPixmap = bitmap->Handle();
	msg.iData.iCreatePixmapCursorReq.iMask = bitmapMask->Handle();
	msg.iData.iCreatePixmapCursorReq.iHotPointX = aHotpointX;
	msg.iData.iCreatePixmapCursorReq.iHotPointY = aHotpointY;
	iVncIpc->SendMessageL(ECreatePixmapCursorReq, msg);

	// Now receive the response
	iVncIpc->RecvMessageL(msg);
	if (msg.iMsgType == ECreatePixmapCursorResp)
		{
		return msg.iData.iCreatePixmapCursorResp.iCursor;
		}
	User::Leave(KErrArgument);
	return Cursor();
	}
	
/**
Move a cursor
@param aX The new x position
@param aY The new y position
*/
void CSymbianWindowSystem::WarpPointerL(int aX, int aY)
	{
	struct vncmsg msg;
	
	// For some reason, the cursor is off by 1 pixel
	msg.iData.iWarpPointer.iCursorX = aX + 1;	
	msg.iData.iWarpPointer.iCursorY = aY + 1;	
	iVncIpc->SendMessageL(EWarpPointer, msg);
	}
	
/**
Free a cursor
@return aCursor The cursor
*/
void CSymbianWindowSystem::FreeCursorL(Cursor aCursor)
	{
	struct vncmsg msg;
	msg.iData.iFreeCursor.iCursor = aCursor;
	iVncIpc->SendMessageL(EFreeCursor, msg);
	}
	
/**
Translate a Unicode character to Key Symbol

Copied from the VNC Viewer GNU Project by Lucas Gomez.

@param aUnicode The unicode character
@return TInt The key sym
*/
TInt CSymbianWindowSystem::TranslateUnicodeToKeysym(TInt aUnicode)
	{
	if((aUnicode>=0x20 && aUnicode<=0x7e) ||
       (aUnicode>=0xa0 && aUnicode<=0xff))
    	return aUnicode;

    TInt min=0;
    TInt max=Keysyms::KKeysymTableSize-1;
    TInt mid;

    while(max>=min) 
	    {
      	mid=(min+max)/2;
      	if(Keysyms::KKeysymTable[mid][1]<aUnicode)
        	min=mid+1;
      	else if(Keysyms::KKeysymTable[mid][1]>aUnicode)
        	max=mid-1;
      	else
        	return Keysyms::KKeysymTable[mid][0];
	    }

    /* no matching Unicode value found */
    return -1;
	}
	
/**
Translate a key code string to a Key Symbol

Copied from the VNC Viewer GNU Project by Lucas Gomez.

@param aKeyCode The keycode string
@return TInt The key sym
*/
KeySym CSymbianWindowSystem::Translate(const char* aKeyCode)
	{
	KeySym sym = NoSymbol;
	
	// 
	TPtrC8 ptrCode((TUint8*)aKeyCode);
	TLex8 lex(ptrCode);
	
	TInt unicodeKey;
	lex.Val(unicodeKey);

	// Raw key	
	if (unicodeKey > 0xFF00)
		return unicodeKey;
	
	TInt val = TranslateUnicodeToKeysym(unicodeKey);
	if (val != -1)
		sym = val;
	return sym;
	}
